#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2022-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import dataclasses
import typing
import uuid

from . import enums


@dataclasses.dataclass(frozen=True)
class AnyType:
    desc_id: uuid.UUID
    name: typing.Optional[str]


@dataclasses.dataclass(frozen=True)
class Element:
    type: AnyType
    cardinality: enums.Cardinality
    is_implicit: bool
    kind: enums.ElementKind


@dataclasses.dataclass(frozen=True)
class SequenceType(AnyType):
    element_type: AnyType


@dataclasses.dataclass(frozen=True)
class SetType(SequenceType):
    pass


@dataclasses.dataclass(frozen=True)
class ObjectType(AnyType):
    elements: typing.Dict[str, Element]


@dataclasses.dataclass(frozen=True)
class BaseScalarType(AnyType):
    pass


@dataclasses.dataclass(frozen=True)
class ScalarType(AnyType):
    base_type: BaseScalarType


@dataclasses.dataclass(frozen=True)
class TupleType(AnyType):
    element_types: typing.Tuple[AnyType]


@dataclasses.dataclass(frozen=True)
class NamedTupleType(AnyType):
    element_types: typing.Dict[str, AnyType]


@dataclasses.dataclass(frozen=True)
class ArrayType(SequenceType):
    pass


@dataclasses.dataclass(frozen=True)
class EnumType(AnyType):
    members: typing.Tuple[str]


@dataclasses.dataclass(frozen=True)
class SparseObjectType(ObjectType):
    pass


@dataclasses.dataclass(frozen=True)
class RangeType(AnyType):
    value_type: AnyType


@dataclasses.dataclass(frozen=True)
class MultiRangeType(AnyType):
    value_type: AnyType



import pathlib
import re
import warnings

from contextlib import contextmanager

from .introspection import get_sql_name, get_mod_and_name
from .introspection import GelORMWarning, FilePrinter


GEL_SCALAR_MAP = {
    'std::bool': ('bool', 'sa.Boolean'),
    'std::str': ('str', 'sa.String'),
    'std::int16': ('int', 'sa.Integer'),
    'std::int32': ('int', 'sa.Integer'),
    'std::int64': ('int', 'sa.Integer'),
    'std::float32': ('float', 'sa.Float'),
    'std::float64': ('float', 'sa.Float'),
    'std::uuid': ('uuid.UUID', 'sa.Uuid'),
    'std::bytes': ('bytes', 'sa.LargeBinary'),
    'std::cal::local_date': ('datetime.date', 'sa.Date'),
    'std::cal::local_time': ('datetime.time', 'sa.Time'),
    'std::cal::local_datetime': ('datetime.datetime', 'sa.DateTime'),
    'std::datetime': ('datetime.datetime', 'sa.TIMESTAMP'),
}

ARRAY_RE = re.compile(r'^array<(?P<el>.+)>$')
NAME_RE = re.compile(r'^(?P<alpha>\w+?)(?P<num>\d*)$')

COMMENT = '''\
#
# Automatically generated from Gel schema.
#
# Do not edit directly as re-generating this file will overwrite any changes.
#\
'''

BASE_STUB = f'''\
{COMMENT}

from sqlalchemy import orm as orm


class Base(orm.DeclarativeBase):
    pass\
'''

MODELS_STUB = f'''\
{COMMENT}

import datetime
import uuid

from typing import List, Optional

import sqlalchemy as sa
from sqlalchemy import orm as orm
'''


def field_name_sort(spec):
    key = spec['name']

    match = NAME_RE.fullmatch(key)
    res = (match.group('alpha'), int(match.group('num') or -1))

    return res


class ModelGenerator(FilePrinter):
    def __init__(self, *, outdir=None, basemodule=None):
        # set the output to be stdout by default, but this is generally
        # expected to be overridden by appropriate files in the `outdir`
        if outdir is not None:
            self.outdir = pathlib.Path(outdir)
        else:
            self.outdir = None

        self.basemodule = basemodule
        super().__init__()

    def init_dir(self, dirpath):
        if not dirpath:
            # nothing to initialize
            return

        path = pathlib.Path(dirpath).resolve()

        # ensure `path` directory exists
        if not path.exists():
            path.mkdir()
        elif not path.is_dir():
            raise NotADirectoryError(
                f'{path!r} exists, but it is not a directory')

        # ensure `path` directory contains `__init__.py`
        (path / '__init__.py').touch()

    def init_sqlabase(self):
        with open(self.outdir / '_sqlabase.py', 'wt') as f:
            self.out = f
            self.write(BASE_STUB)

    @contextmanager
    def init_module(self, mod, modules):
        if any(m.startswith(f'{mod}::') for m in modules):
            # This is a prefix in another module, thus it is part of a nested
            # module structure.
            dirpath = mod.split('::')
            filename = '__init__.py'
        else:
            # This is a leaf module, so we just need to create a corresponding
            # <mod>.py file.
            *dirpath, filename = mod.split('::')
            filename = f'{filename}.py'

        # Along the dirpath we need to ensure that all packages are created
        path = self.outdir
        for el in dirpath:
            path = path / el
            self.init_dir(path)

        with open(path / filename, 'wt') as f:
            try:
                self.out = f
                self.write(MODELS_STUB)
                relimport = '.' * len(dirpath)
                self.write(f'from {relimport}._sqlabase import Base')
                self.write(f'from {relimport}._tables import *')
                yield f
            finally:
                self.out = None

    def get_fk(self, mod, table, curmod):
        if mod == curmod:
            # No need for anything fancy within the same schema
            return f'sa.ForeignKey("{table}.id")'
        else:
            return f'sa.ForeignKey("{mod}.{table}.id")'

    def get_py_name(self, mod, name, curmod):
        if False and mod == curmod:
            # No need for anything fancy within the same module
            return f"'{name}'"
        else:
            mod = mod.replace('::', '.')
            return f"'{self.basemodule}.{mod}.{name}'"

    def spec_to_modules_dict(self, spec):
        modules = {
            mod: {} for mod in sorted(spec['modules'])
        }

        if len(spec['prop_objects']) > 0:
            warnings.warn(
                f"Skipping multi properties: SQLAlchemy reflection doesn't "
                f"support multi properties as they produce models without a "
                f"clear identity.",
                GelORMWarning,
            )

        for rec in spec['link_tables']:
            mod = rec['module']
            if 'link_tables' not in modules[mod]:
                modules[mod]['link_tables'] = []
            modules[mod]['link_tables'].append(rec)

        for lobj in spec['link_objects']:
            mod = lobj['module']
            if 'link_objects' not in modules[mod]:
                modules[mod]['link_objects'] = {}
            modules[mod]['link_objects'][lobj['name']] = lobj

        for rec in spec['object_types']:
            mod, name = get_mod_and_name(rec['name'])
            if 'object_types' not in modules[mod]:
                modules[mod]['object_types'] = {}
            modules[mod]['object_types'][name] = rec

        return modules

    def render_models(self, spec):
        # The modules dict will be populated with the respective types, link
        # tables, etc., since they will need to be put in their own files. We
        # sort the modules so that nested modules are initialized from root to
        # leaf.
        modules = self.spec_to_modules_dict(spec)

        # Initialize the base directory
        self.init_dir(self.outdir)
        self.init_sqlabase()

        with open(self.outdir / '_tables.py', 'wt') as f:
            self.out = f
            self.write(MODELS_STUB)
            self.write(f'from ._sqlabase import Base')

            link_tables = sorted(spec['link_tables'], key=lambda x: x['name'])
            for rec in link_tables:
                self.write()
                self.render_link_table(rec)

        for mod, maps in modules.items():
            with self.init_module(mod, modules):
                if not maps:
                    # skip apparently empty modules
                    continue

                link_objects = sorted(
                    maps.get('link_objects', {}).values(),
                    key=lambda x: x['name'],
                )
                for lobj in link_objects:
                    self.write()
                    self.render_link_object(lobj, modules)

                object_types = sorted(
                    maps.get('object_types', {}).values(),
                    key=lambda x: x['name'],
                )
                for rec in object_types:
                    self.write()
                    self.render_type(rec, modules)

    def render_link_table(self, spec):
        mod, source = get_mod_and_name(spec["source"])
        tmod, target = get_mod_and_name(spec["target"])
        s_fk = self.get_fk(mod, source, 'default')
        t_fk = self.get_fk(tmod, target, 'default')

        self.write()
        self.write(f'{spec["name"]} = sa.Table(')
        self.indent()
        self.write(f'{spec["table"]!r},')
        self.write(f'Base.metadata,')
        # source is in the same module as this table
        self.write(f'sa.Column("source", {s_fk}),')
        self.write(f'sa.Column("target", {t_fk}),')
        self.write(f'schema={mod!r},')
        self.dedent()
        self.write(f')')

    def render_link_object(self, spec, modules):
        mod = spec['module']
        name = spec['name']
        sql_name = spec['table']
        source_name, source_link = sql_name.split('.')

        self.write()
        self.write(f'class {name}(Base):')
        self.indent()
        self.write(f'__tablename__ = {sql_name!r}')
        if mod != 'default':
            self.write(f'__table_args__ = {{"schema": {mod!r}}}')
        # We rely on Gel for maintaining integrity and various on delete
        # triggers, so the rows may be deleted in a different way from what
        # SQLAlchemy expects.
        self.write('__mapper_args__ = {"confirm_deleted_rows": False}')
        self.write()

        # No ids for these intermediate objects
        if spec['links']:
            self.write()
            self.write('# Links:')

            for link in spec['links']:
                lname = link['name']
                tmod, target = get_mod_and_name(link['target']['name'])
                fk = self.get_fk(tmod, target, mod)
                pyname = self.get_py_name(tmod, target, mod)
                self.write(f'{lname}_id: orm.Mapped[uuid.UUID] = orm.mapped_column(')
                self.indent()
                self.write(f'{lname!r},')
                self.write(f'sa.Uuid(),')
                self.write(f'{fk},')
                self.write(f'primary_key=True,')
                self.write(f'nullable=False,')
                self.dedent()
                self.write(')')

                if lname == 'source':
                    bklink = source_link
                else:
                    src = modules[mod]['object_types'][source_name]
                    bklink = f'_{source_link}_{source_name}'

                self.write(
                    f'{lname}: orm.Mapped[{pyname}] = '
                    f'orm.relationship(back_populates={bklink!r})'
                )

        if spec['properties']:
            self.write()
            self.write('# Properties:')

            for prop in spec['properties']:
                if prop['name'] != 'id':
                    self.render_prop(prop, mod, name, {})

        self.dedent()

    def render_type(self, spec, modules):
        # assume nice names for now
        mod, name = get_mod_and_name(spec['name'])
        sql_name = get_sql_name(spec['name'])

        self.write()
        self.write(f'class {name}(Base):')
        self.indent()
        self.write(f'__tablename__ = {sql_name!r}')
        if mod != 'default':
            self.write(f'__table_args__ = {{"schema": {mod!r}}}')
        # We rely on Gel for maintaining integrity and various on delete
        # triggers, so the rows may be deleted in a different way from what
        # SQLAlchemy expects.
        self.write('__mapper_args__ = {"confirm_deleted_rows": False}')
        self.write()

        # Add two fields that all objects have
        self.write(f'id: orm.Mapped[uuid.UUID] = orm.mapped_column(')
        self.indent()
        self.write(f"sa.Uuid(),")
        self.write(f"primary_key=True,")
        self.write(f"server_default='uuid_generate_v4()',")
        self.dedent()
        self.write(f')')

        # This is maintained entirely by Gel, the server_default simply
        # indicates to SQLAlchemy that this value may be omitted.
        self.write(f'gel_type_id: orm.Mapped[uuid.UUID] = orm.mapped_column(')
        self.indent()
        self.write(f"'__type__',")
        self.write(f"sa.Uuid(),")
        self.write(f"server_default='PLACEHOLDER',")
        self.dedent()
        self.write(f")")

        if spec['properties']:
            self.write()
            self.write('# Properties:')

            properties = sorted(spec['properties'], key=field_name_sort)
            for prop in properties:
                if prop['name'] != 'id':
                    self.render_prop(prop, mod, name, modules)

        if spec['links']:
            self.write()
            self.write('# Links:')

            links = sorted(spec['links'], key=field_name_sort)
            for link in links:
                self.render_link(link, mod, name, modules)

        if spec['backlinks']:
            self.write()
            self.write('# Back-links:')

            backlinks = sorted(spec['backlinks'], key=field_name_sort)
            for link in backlinks:
                self.render_backlink(link, mod, modules)

        self.dedent()

    def render_prop(self, spec, mod, parent, modules, *, is_pk=False):
        name = spec['name']
        nullable = not spec['required']
        cardinality = spec['cardinality']

        target = spec['target']['name']
        is_array = False
        match = ARRAY_RE.fullmatch(target)
        if match:
            is_array = True
            target = match.group('el')

        try:
            pytype, sqlatype = GEL_SCALAR_MAP[target]
            sqlatype = sqlatype + '()'
        except KeyError:
            warnings.warn(
                f'Scalar type {target} is not supported',
                GelORMWarning,
            )
            # Skip rendering this one
            return

        if is_array:
            pytype = f'List[{pytype}]'
            sqlatype = f'sa.ARRAY({sqlatype})'

        if is_pk:
            # special case of a primary key property (should only happen to
            # 'target' in multi property table)
            self.write(f'{name}: orm.Mapped[{pytype}] = orm.mapped_column(')
            self.indent()
            self.write(f'{sqlatype}, primary_key=True, nullable=False,')
            self.dedent()
            self.write(f')')
        elif cardinality == 'Many':
            # skip it
            return

        else:
            # plain property
            self.write(f'{name}: orm.Mapped[{pytype}] = orm.mapped_column(')
            self.indent()
            self.write(f'{sqlatype}, nullable={nullable},')
            self.dedent()
            self.write(f')')

    def render_link(self, spec, mod, parent, modules):
        name = spec['name']
        nullable = not spec['required']
        tmod, target = get_mod_and_name(spec['target']['name'])
        source = modules[mod]['object_types'][parent]
        cardinality = spec['cardinality']
        bklink = f'_{name}_{parent}'

        if spec.get('has_link_object'):
            # intermediate object will have the actual source and target
            # links, so the link here needs to be treated similar to a
            # back-link.
            linkobj = modules[mod]['link_objects'][f'{parent}_{name}_link']
            target = linkobj['name']
            tmod = linkobj['module']
            pyname = self.get_py_name(tmod, target, mod)

            if cardinality == 'One':
                self.write(
                    f'{name}: orm.Mapped[{pyname}] = '
                    f"orm.relationship(back_populates='source')"
                )
            elif cardinality == 'Many':
                self.write(
                    f'{name}: orm.Mapped[List[{pyname}]] = '
                    f"orm.relationship(back_populates='source')"
                )

            if cardinality == 'One':
                tmap = f'orm.Mapped[{pyname}]'
            elif cardinality == 'Many':
                tmap = f'orm.Mapped[List[{pyname}]]'
            # We want the cascade to delete orphans here as the intermediate
            # objects represent links and must not exist without source.
            self.write(f'{name}: {tmap} = orm.relationship(')
            self.indent()
            self.write(f"back_populates='source',")
            self.write(f"cascade='all, delete-orphan',")
            self.dedent()
            self.write(')')

        else:
            fk = self.get_fk(tmod, target, mod)
            pyname = self.get_py_name(tmod, target, mod)

            if cardinality == 'One':
                self.write(
                    f'{name}_id: orm.Mapped[uuid.UUID] = orm.mapped_column(')
                self.indent()
                self.write(f'sa.Uuid(), {fk}, nullable={nullable},')
                self.dedent()
                self.write(f')')

                self.write(f'{name}: orm.Mapped[{pyname}] = orm.relationship(')
                self.indent()
                self.write(f'back_populates={bklink!r},')
                self.dedent()
                self.write(f')')

            elif cardinality == 'Many':
                secondary = f'{parent}_{name}_table'

                self.write(
                    f'{name}: orm.Mapped[List[{pyname}]] = orm.relationship(')
                self.indent()
                self.write(f'{pyname},')
                self.write(f'secondary={secondary},')
                self.write(f'back_populates={bklink!r},')
                self.dedent()
                self.write(f')')

    def render_backlink(self, spec, mod, modules):
        name = spec['name']
        tmod, target = get_mod_and_name(spec['target']['name'])
        cardinality = spec['cardinality']
        exclusive = spec['exclusive']
        bklink = spec['fwname']

        if spec.get('has_link_object'):
            # intermediate object will have the actual source and target
            # links, so the link here needs to refer to the intermediate
            # object and 'target' as back-link.
            linkobj = modules[tmod]['link_objects'][f'{target}_{bklink}_link']
            target = linkobj['name']
            tmod = linkobj['module']
            pyname = self.get_py_name(tmod, target, mod)

            if cardinality == 'One':
                tmap = f'orm.Mapped[{pyname}]'
            elif cardinality == 'Many':
                tmap = f'orm.Mapped[List[{pyname}]]'
            # We want the cascade to delete orphans here as the intermediate
            # objects represent links and must not exist without target.
            self.write(f'{name}: {tmap} = orm.relationship(')
            self.indent()
            self.write(f"back_populates='target',")
            self.write(f"cascade='all, delete-orphan',")
            self.dedent()
            self.write(')')

        else:
            pyname = self.get_py_name(tmod, target, mod)
            if exclusive:
                # This is a backlink from a single link. There is no link table
                # involved.
                if cardinality == 'One':
                    self.write(f'{name}: orm.Mapped[{pyname}] = \\')
                    self.indent()
                    self.write(f'orm.relationship(back_populates={bklink!r})')
                    self.dedent()

                elif cardinality == 'Many':
                    self.write(f'{name}: orm.Mapped[List[{pyname}]] = \\')
                    self.indent()
                    self.write(f'orm.relationship(back_populates={bklink!r})')
                    self.dedent()

            else:
                # This backlink involves a link table, so we still treat it as
                # a Many-to-Many.
                secondary = f'{target}_{bklink}_table'

                self.write(f'{name}: orm.Mapped[List[{pyname}]] = \\')
                self.indent()
                self.write(f'orm.relationship(')
                self.indent()
                self.write(f'{pyname},')
                self.write(f'secondary={secondary},')
                self.write(f'back_populates={bklink!r},')
                self.dedent()
                self.write(')')
                self.dedent()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import enum


@cython.final
cdef class EnumCodec(BaseCodec):

    cdef encode(self, WriteBuffer buf, object obj):
        if not isinstance(obj, (self.cls, str)):
            try:
                obj = self.cls._try_from(obj)
            except (TypeError, ValueError):
                raise TypeError(
                    f'a str or gel.EnumValue(__tid__={self.cls.__tid__}) '
                    f'is expected as a valid enum argument, '
                    f'got {type(obj).__name__}') from None
        pgproto.text_encode(DEFAULT_CODEC_CONTEXT, buf, str(obj))

    cdef decode(self, FRBuffer *buf):
        label = pgproto.text_decode(DEFAULT_CODEC_CONTEXT, buf)
        return self.cls(label)

    @staticmethod
    cdef BaseCodec new(bytes tid, tuple enum_labels):
        cdef:
            EnumCodec codec

        codec = EnumCodec.__new__(EnumCodec)

        codec.tid = tid
        codec.name = 'Enum'
        cls = "DerivedEnumValue"
        bases = (datatypes.EnumValue,)
        classdict = enum.EnumMeta.__prepare__(cls, bases)
        classdict["__module__"] = "gel"
        classdict["__qualname__"] = "gel.DerivedEnumValue"
        classdict["__tid__"] = pgproto.UUID(tid)
        for label in enum_labels:
            classdict[label.upper()] = label
        codec.cls = enum.EnumMeta(cls, bases, classdict)
        for index, label in enumerate(enum_labels):
            codec.cls(label)._index_ = index

        return codec

    def make_type(self, describe_context):
        return describe.EnumType(
            desc_id=uuid.UUID(bytes=self.tid),
            name=self.type_name,
            members=tuple(x.value for x in self.cls.__members__.values())
        )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import contextlib
import json
import os
import pathlib
import re
import sys
import tempfile
import unittest
import warnings
from unittest import mock


from gel import con_utils
from edgedb import errors


class TestConUtils(unittest.TestCase):
    maxDiff = 1000

    error_mapping = {
        'credentials_file_not_found': (
            RuntimeError, 'cannot read credentials'),
        'project_not_initialised': (
            errors.ClientConnectionError,
            r'Found `\w+.toml` but the project is not initialized'),
        'no_options_or_toml': (
            errors.ClientConnectionError,
            'no `gel.toml` found and no connection options specified'),
        'invalid_credentials_file': (
            RuntimeError, 'cannot read credentials'),
        'invalid_dsn_or_instance_name': (
            ValueError, 'invalid DSN or instance name'),
        'invalid_instance_name': (
            ValueError, 'invalid instance name'),
        'invalid_dsn': (ValueError, 'invalid DSN'),
        'unix_socket_unsupported': (
            ValueError, 'unix socket paths not supported'),
        'invalid_host': (ValueError, 'invalid host'),
        'invalid_port': (ValueError, 'invalid port'),
        'invalid_user': (ValueError, 'invalid user'),
        'invalid_database': (ValueError, 'invalid database'),
        'multiple_compound_env': (
            errors.ClientConnectionError,
            'Cannot have more than one of the following connection '
            + 'environment variables'),
        'multiple_compound_opts': (
            errors.ClientConnectionError,
            'Cannot have more than one of the following connection options'),
        'exclusive_options': (
            errors.ClientConnectionError,
            'are mutually exclusive'),
        'env_not_found': (
            ValueError, 'environment variable ".*" doesn\'t exist'),
        'file_not_found': (FileNotFoundError, 'No such file or directory'),
        'invalid_tls_security': (
            ValueError, 'tls_security can only be one of `insecure`, '
            '|tls_security must be set to strict'),
        'invalid_secret_key': (
            errors.ClientConnectionError, "Invalid secret key"),
        'secret_key_not_found': (
            errors.ClientConnectionError,
            "Cannot connect to cloud instances without secret key"),
        'docker_tcp_port': (
            'EDGEDB_PORT in "tcp://host:port" format, so will be ignored'
        ),
        'gel_and_edgedb': (
            r'Both GEL_\w+ and EDGEDB_\w+ are set; EDGEDB_\w+ will be ignored'
        ),
    }

    @contextlib.contextmanager
    def environ(self, **kwargs):
        old_environ = os.environ.copy()
        try:
            # Apparently assigning to os.environ behaves weirdly, so
            # we clear and update.
            os.environ.clear()
            os.environ.update(kwargs)
            yield
        finally:
            os.environ.clear()
            os.environ.update(old_environ)

    def run_testcase(self, testcase):
        test_env = testcase.get('env', {})

        fs = testcase.get('fs')

        opts = testcase.get('opts', {})
        dsn = opts['instance'] if 'instance' in opts else opts.get('dsn')
        credentials = opts.get('credentials')
        credentials_file = opts.get('credentialsFile')
        host = opts.get('host')
        port = opts.get('port')
        database = opts.get('database')
        branch = opts.get('branch')
        user = opts.get('user')
        password = opts.get('password')
        secret_key = opts.get('secretKey')
        tls_ca = opts.get('tlsCA')
        tls_ca_file = opts.get('tlsCAFile')
        tls_security = opts.get('tlsSecurity')
        tls_server_name = opts.get('tlsServerName')
        server_settings = opts.get('serverSettings')
        wait_until_available = opts.get('waitUntilAvailable')

        other_opts = testcase.get('other_opts', {})
        timeout = other_opts.get('timeout')
        command_timeout = other_opts.get('command_timeout')

        expected = testcase.get('result')
        expected_error = testcase.get('error')
        if expected_error and expected_error.get('type'):
            expected_error = self.error_mapping.get(expected_error.get('type'))
            if not expected_error:
                raise RuntimeError(
                    f"unknown error type: {testcase.get('error').get('type')}")

        if expected is None and expected_error is None:
            raise RuntimeError(
                'invalid test case: either "result" or "error" key '
                'has to be specified')
        if expected is not None and expected_error is not None:
            raise RuntimeError(
                'invalid test case: either "result" or "error" key '
                'has to be specified, got both')

        result = None
        with contextlib.ExitStack() as es:
            es.enter_context(self.subTest(dsn=dsn, env=test_env, opts=opts))
            es.enter_context(self.environ(**test_env))

            stat_result = os.stat(os.getcwd())
            es.enter_context(
                mock.patch('os.stat', lambda _, **__: stat_result)
            )

            if fs:
                cwd = fs.get('cwd')
                homedir = fs.get('homedir')
                files = fs.get('files')

                if cwd:
                    es.enter_context(mock.patch('os.getcwd', lambda: cwd))
                if homedir:
                    homedir = pathlib.Path(homedir)
                    es.enter_context(
                        mock.patch('pathlib.Path.home', lambda: homedir)
                    )
                if files:
                    for f, v in files.copy().items():
                        if "${HASH}" in f:
                            hash = con_utils._hash_path(v['project-path'])
                            dir = f.replace("${HASH}", hash)
                            files[dir] = ""
                            instance = os.path.join(dir, 'instance-name')
                            files[instance] = v['instance-name']
                            project = os.path.join(dir, 'project-path')
                            files[project] = v['project-path']
                            if 'cloud-profile' in v:
                                profile = os.path.join(dir, 'cloud-profile')
                                files[profile] = v['cloud-profile']
                            if 'database' in v:
                                database_file = os.path.join(dir, 'database')
                                files[database_file] = v['database']
                            del files[f]

                    es.enter_context(
                        mock.patch(
                            'os.path.exists',
                            lambda filepath: str(filepath) in files
                        )
                    )
                    es.enter_context(
                        mock.patch(
                            'os.path.isfile',
                            lambda filepath: str(filepath) in files
                        )
                    )

                    es.enter_context(mock.patch(
                        'os.stat',
                        lambda d, **_: mock.Mock(st_dev=0),
                    ))

                    es.enter_context(
                        mock.patch('os.path.realpath', lambda f: f)
                    )

                    def mocked_open(filepath, *args, **kwargs):
                        if str(filepath) in files:
                            return mock.mock_open(
                                read_data=files.get(str(filepath))
                            )()
                        raise FileNotFoundError(
                            f"[Errno 2] No such file or directory: " +
                            f"'{filepath}'"
                        )
                    es.enter_context(mock.patch('builtins.open', mocked_open))

            if expected_error:
                es.enter_context(self.assertRaisesRegex(*expected_error))

            warning_list = es.enter_context(
                warnings.catch_warnings(record=True)
            )

            connect_config, client_config = con_utils.parse_connect_arguments(
                dsn=dsn,
                host=host,
                port=port,
                credentials=credentials,
                credentials_file=credentials_file,
                database=database,
                branch=branch,
                user=user,
                password=password,
                secret_key=secret_key,
                tls_ca=tls_ca,
                tls_ca_file=tls_ca_file,
                tls_security=tls_security,
                tls_server_name=tls_server_name,
                timeout=timeout,
                command_timeout=command_timeout,
                server_settings=server_settings,
                wait_until_available=wait_until_available,
            )

            result = {
                'address': [
                    connect_config.address[0], connect_config.address[1]
                ],
                'database': connect_config.database,
                'branch': connect_config.branch,
                'user': connect_config.user,
                'password': connect_config.password,
                'secretKey': connect_config.secret_key,
                'tlsCAData': connect_config._tls_ca_data,
                'tlsSecurity': connect_config.tls_security,
                'serverSettings': connect_config.server_settings,
                'waitUntilAvailable': float(
                    client_config.wait_until_available
                ),
                'tlsServerName': connect_config.tls_server_name,
            }

        if expected is not None:
            self.assertEqual(expected, result)

        expected_warnings = testcase.get('warnings', [])
        self.assertEqual(len(expected_warnings), len(warning_list))
        for expected_key, warning in zip(expected_warnings, warning_list):
            expected_warning = self.error_mapping.get(expected_key)
            if not expected_warning:
                raise RuntimeError(
                    f"unknown error type: {expected_key}"
                )
            if not re.match(expected_warning, str(warning.message)):
                raise AssertionError(
                    f'Warning "{warning.message}" does not match '
                    f'"{expected_warning}"'
                )

    def test_test_connect_params_environ(self):
        self.assertNotIn('AAAAAAAAAA123', os.environ)
        self.assertNotIn('AAAAAAAAAA456', os.environ)
        self.assertNotIn('AAAAAAAAAA789', os.environ)

        try:

            os.environ['AAAAAAAAAA456'] = '123'
            os.environ['AAAAAAAAAA789'] = '123'

            with self.environ(AAAAAAAAAA123='1',
                              AAAAAAAAAA456='2'):

                self.assertEqual(os.environ['AAAAAAAAAA123'], '1')
                self.assertEqual(os.environ['AAAAAAAAAA456'], '2')
                self.assertNotIn('AAAAAAAAAA789', os.environ)

            self.assertNotIn('AAAAAAAAAA123', os.environ)
            self.assertEqual(os.environ['AAAAAAAAAA456'], '123')
            self.assertEqual(os.environ['AAAAAAAAAA789'], '123')

        finally:
            for key in {'AAAAAAAAAA123', 'AAAAAAAAAA456', 'AAAAAAAAAA789'}:
                if key in os.environ:
                    del os.environ[key]

    def test_test_connect_params_run_testcase_01(self):
        with self.environ(EDGEDB_PORT='777'):
            self.run_testcase({
                'env': {
                    'EDGEDB_HOST': 'abc'
                },
                'opts': {
                    'user': '__test__',
                },
                'result': {
                    'address': ['abc', 5656],
                    'database': 'edgedb',
                    'branch': '__default__',
                    'user': '__test__',
                    'password': None,
                    'secretKey': None,
                    'tlsCAData': None,
                    'tlsSecurity': 'strict',
                    'serverSettings': {},
                    'waitUntilAvailable': 30,
                    'tlsServerName': None,
                },
            })

    def test_test_connect_params_run_testcase_02(self):
        with self.environ(EDGEDB_PORT='777'):
            self.run_testcase({
                'env': {
                    'EDGEDB_HOST': 'abc'
                },
                'opts': {
                    'user': '__test__',
                    'branch': 'new_branch',
                },
                'result': {
                    'address': ['abc', 5656],
                    'database': 'new_branch',
                    'branch': 'new_branch',
                    'user': '__test__',
                    'password': None,
                    'secretKey': None,
                    'tlsCAData': None,
                    'tlsSecurity': 'strict',
                    'serverSettings': {},
                    'waitUntilAvailable': 30,
                    'tlsServerName': None,
                },
            })

    def test_connect_params(self):
        testcases_path = (
            pathlib.Path(__file__).parent
            / "shared-client-testcases"
            / "connection_testcases.json"
        )
        try:
            with open(testcases_path, encoding="utf-8") as f:
                testcases = json.load(f)
        except FileNotFoundError as err:
            raise FileNotFoundError(
                f'Failed to read "connection_testcases.json": {err}.\n' +
                f'Is the "shared-client-testcases" submodule initialised? ' +
                f'Try running "git submodule update --init".'
            )

        for i, testcase in enumerate(testcases):
            with self.subTest(i=i, name=testcase.get('name')):
                wait_until_available = \
                    testcase.get('result', {}).get('waitUntilAvailable')
                if wait_until_available:
                    testcase['result']['waitUntilAvailable'] = \
                        con_utils._validate_wait_until_available(
                            wait_until_available)
                platform = testcase.get('platform')
                if testcase.get('fs') and (
                    sys.platform == 'win32' or platform == 'windows'
                    or (platform is None and sys.platform == 'darwin')
                    or (platform == 'macos' and sys.platform != 'darwin')
                ):
                    continue

                self.run_testcase(testcase)

    @mock.patch("gel.platform.config_dir",
                lambda: pathlib.Path("/home/user/.config/edgedb"))
    @mock.patch("gel.platform.IS_WINDOWS", False)
    @mock.patch("pathlib.Path.exists", lambda p: True)
    @mock.patch("os.path.realpath", lambda p: p)
    def test_stash_path(self):
        self.assertEqual(
            con_utils._stash_path("/home/user/work/project1"),
            pathlib.Path("/home/user/.config/edgedb/projects/project1-"
                         "cf1c841351bf7f147d70dcb6203441cf77a05249"),
        )

    def test_project_config(self):
        with tempfile.TemporaryDirectory() as tmp:
            tmp = os.path.realpath(tmp)
            base = pathlib.Path(tmp)
            home = base / 'home'
            project = base / 'project'
            projects = home / '.edgedb' / 'projects'
            creds = home / '.edgedb' / 'credentials'
            os.makedirs(projects)
            os.makedirs(creds)
            os.makedirs(project)
            with open(project / 'edgedb.toml', 'wt') as f:
                f.write('')  # app don't read toml file
            with open(creds / 'inst1.json', 'wt') as f:
                f.write(json.dumps({
                    "host": "inst1.example.org",
                    "port": 12323,
                    "user": "inst1_user",
                    "password": "passw1",
                    "database": "inst1_db",
                }))

            with mock.patch('gel.platform.config_dir',
                            lambda: home / '.edgedb'), \
                    mock.patch('os.getcwd', lambda: str(project)):
                stash_path = con_utils._stash_path(project)
                instance_file = stash_path / 'instance-name'
                os.makedirs(stash_path)
                with open(instance_file, 'wt') as f:
                    f.write('inst1')

                connect_config, client_config = (
                    con_utils.parse_connect_arguments(
                        dsn=None,
                        host=None,
                        port=None,
                        credentials=None,
                        credentials_file=None,
                        user=None,
                        password=None,
                        secret_key=None,
                        database=None,
                        branch=None,
                        tls_ca=None,
                        tls_ca_file=None,
                        tls_security=None,
                        tls_server_name=None,
                        timeout=10,
                        command_timeout=None,
                        server_settings=None,
                        wait_until_available=30,
                    )
                )

        self.assertEqual(connect_config.address, ('inst1.example.org', 12323))
        self.assertEqual(connect_config.user, 'inst1_user')
        self.assertEqual(connect_config.password, 'passw1')
        self.assertEqual(connect_config.database, 'inst1_db')

    def test_validate_wait_until_available(self):
        invalid = [
            ' ',
            ' PT1S',
            '',
            '-.1 s',
            '-.1s',
            '-.5 second',
            '-.5 seconds',
            '-.5second',
            '-.5seconds',
            '-.s',
            '-1.s',
            '.s',
            '.seconds',
            '1.s',
            '1h-120m3600s',
            '1hour-120minute3600second',
            '1hours-120minutes3600seconds',
            '1hours120minutes3600seconds',
            '2.0hour46.0minutes39.0seconds',
            '2.0hours46.0minutes39.0seconds',
            '20 hours with other stuff should not be valid',
            '20 minutes with other stuff should not be valid',
            '20 ms with other stuff should not be valid',
            '20 seconds with other stuff should not be valid',
            '20 us with other stuff should not be valid',
            '2hour46minute39second',
            '2hours46minutes39seconds',
            '3 hours is longer than 10 seconds',
            'P-.D',
            'P-D',
            'PD',
            'PT.S',
            'PT1S ',
            '\t',
            'not a duration',
            's',
        ]

        for string in invalid:
            with self.subTest(string):
                with self.assertRaises(ValueError):
                    con_utils._validate_wait_until_available(string)

        valid = [
            (" 1s ", 1),
            (" 1s", 1),
            ("-0s", 0),
            ("-1.0h", -3600),
            ("-1.0hour", -3600),
            ("-1.0hours", -3600),
            ("-1.0m", -60),
            ("-1.0minute", -60),
            ("-1.0minutes", -60),
            ("-1.0ms", -0.001),
            ("-1.0s", -1),
            ("-1.0second", -1),
            ("-1.0seconds", -1),
            ("-1.0us", -0.000001),
            ("-1h", -3600),
            ("-1hour", -3600),
            ("-1hours", -3600),
            ("-1m", -60),
            ("-1minute", -60),
            ("-1minutes", -60),
            ("-1ms", -0.001),
            ("-1s", -1),
            ("-1second", -1),
            ("-1seconds", -1),
            ("-1us", -0.000001),
            ("-2h 60m 3600s", 0),
            ("-\t2\thour\t60\tminute\t3600\tsecond", 0),
            (".1h", 360),
            (".1hour", 360),
            (".1hours", 360),
            (".1m", 6),
            (".1minute", 6),
            (".1minutes", 6),
            (".1ms", 0.0001),
            (".1s", 0.1),
            (".1second", 0.1),
            (".1seconds", 0.1),
            (".1us", 0.0000001),
            ("1   hour 60  minute -   7200   second", 0),
            ("1   hours 60  minutes -   7200   seconds", 0),
            ("1.0h", 3600),
            ("1.0hour", 3600),
            ("1.0hours", 3600),
            ("1.0m", 60),
            ("1.0minute", 60),
            ("1.0minutes", 60),
            ("1.0ms", 0.001),
            ("1.0s", 1),
            ("1.0second", 1),
            ("1.0seconds", 1),
            ("1.0us", 0.000001),
            ("1h -120m 3600s", 0),
            ("1h -120m3600s", 0),
            ("1h 60m -7200s", 0),
            ("1h", 3600),
            ("1hour", 3600),
            ("1hours -120minutes 3600seconds", 0),
            ("1hours", 3600),
            ("1m", 60),
            ("1minute", 60),
            ("1minutes", 60),
            ("1ms", 0.001),
            ("1s ", 1),
            ("1s", 1),
            ("1s\t", 1),
            ("1second", 1),
            ("1seconds", 1),
            ("1us", 0.000001),
            ("2  h  46  m  39  s", 9999),
            ("2  hour  46  minute  39  second", 9999),
            ("2  hours  46  minutes  39  seconds", 9999),
            ("2.0  h  46.0  m  39.0  s", 9999),
            ("2.0  hour  46.0  minute  39.0  second", 9999),
            ("2.0  hours  46.0  minutes  39.0  seconds", 9999),
            ("2.0h 46.0m 39.0s", 9999),
            ("2.0h46.0m39.0s", 9999),
            ("2.0hour 46.0minute 39.0second", 9999),
            ("2.0hours 46.0minutes 39.0seconds", 9999),
            ("2h 46m 39s", 9999),
            ("2h46m39s", 9999),
            ("2hour 46minute 39second", 9999),
            ("2hours 46minutes 39seconds", 9999),
            ("39.0\tsecond 2.0  hour  46.0  minute", 9999),
            ("PT", 0),
            ("PT-.1", -360),
            ("PT-.1H", -360),
            ("PT-.1M", -6),
            ("PT-.1S", -0.1),
            ("PT-0.000001S", -0.000001),
            ("PT-0S", 0),
            ("PT-1", -3600),
            ("PT-1.", -3600),
            ("PT-1.0", -3600),
            ("PT-1.0H", -3600),
            ("PT-1.0M", -60),
            ("PT-1.0S", -1),
            ("PT-1.H", -3600),
            ("PT-1.M", -60),
            ("PT-1.S", -1),
            ("PT-1H", -3600),
            ("PT-1M", -60),
            ("PT-1S", -1),
            ("PT.1", 360),
            ("PT.1H", 360),
            ("PT.1M", 6),
            ("PT.1S", 0.1),
            ("PT0.000001S", 0.000001),
            ("PT0S", 0),
            ("PT1", 3600),
            ("PT1.", 3600),
            ("PT1.0", 3600),
            ("PT1.0H", 3600),
            ("PT1.0M", 60),
            ("PT1.0M", 60),
            ("PT1.0S", 1),
            ("PT1.H", 3600),
            ("PT1.M", 60),
            ("PT1.S", 1),
            ("PT1H", 3600),
            ("PT1M", 60),
            ("PT1S", 1),
            ("PT2.0H46.0M39.0S", 9999),
            ("PT2H46M39S", 9999),
            ("\t-\t2\thours\t60\tminutes\t3600\tseconds\t", 0),
            ("\t1s", 1),
            ("\t1s\t", 1),
        ]
        for string, expected in valid:
            with self.subTest(duration=string):
                self.assertEqual(
                    expected,
                    con_utils._validate_wait_until_available(string)
                )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import pathlib
import sys
import unittest


def suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover(str(pathlib.Path(__file__).parent),
                                      pattern='test_*.py')
    return test_suite


if __name__ == '__main__':
    runner = unittest.runner.TextTestRunner(verbosity=2)
    result = runner.run(suite())
    sys.exit(not result.wasSuccessful())



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import pathlib
import shutil
import subprocess
import os
import tempfile

from gel import _testbase as tb


ASSERT_SUFFIX = os.environ.get("EDGEDB_TEST_CODEGEN_ASSERT_SUFFIX", ".assert")


class TestCodegen(tb.AsyncQueryTestCase):
    SETUP = '''
        create extension pgvector;
        create scalar type v3 extending ext::pgvector::vector<3>;
    '''

    TEARDOWN = '''
        drop scalar type v3;
        drop extension pgvector;
    '''

    async def test_codegen(self):
        env = os.environ.copy()
        env.update(
            {
                f"EDGEDB_{k.upper()}": str(v)
                for k, v in self.get_connect_args().items()
            }
        )
        env["EDGEDB_DATABASE"] = self.get_database_name()
        container = pathlib.Path(__file__).absolute().parent / "codegen"
        with tempfile.TemporaryDirectory() as td:
            td_path = pathlib.Path(td)
            shutil.copytree(container / "linked", td_path / "linked")
            for project in container.iterdir():
                if project.name == "linked":
                    continue
                with self.subTest(msg=project.name):
                    cwd = td_path / project.name
                    shutil.copytree(project, cwd)
                    try:
                        await self._test_codegen(env, cwd)
                    except subprocess.CalledProcessError as e:
                        self.fail("Codegen failed: " + e.stdout.decode())

    async def _test_codegen(self, env, cwd: pathlib.Path):
        async def run(*args, extra_env=None):
            if extra_env is None:
                env_ = env
            else:
                env_ = env.copy()
                env_.update(extra_env)
            p = await asyncio.create_subprocess_exec(
                *args,
                cwd=cwd,
                env=env_,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            try:
                await asyncio.wait_for(p.wait(), 120)
            except asyncio.TimeoutError:
                p.terminate()
                await p.wait()
                raise
            else:
                if p.returncode:
                    raise subprocess.CalledProcessError(
                        p.returncode, args, output=await p.stdout.read(),
                    )

        cmd = env.get("EDGEDB_PYTHON_TEST_CODEGEN_CMD", "gel-py")
        await run(
            cmd, extra_env={"EDGEDB_PYTHON_CODEGEN_PY_VER": "3.8.5"}
        )
        await run(
            cmd,
            "--target",
            "blocking",
            "--no-skip-pydantic-validation",
            extra_env={"EDGEDB_PYTHON_CODEGEN_PY_VER": "3.9.2"},
        )
        await run(
            cmd,
            "--target",
            "async",
            "--file",
            "--no-skip-pydantic-validation",
            extra_env={"EDGEDB_PYTHON_CODEGEN_PY_VER": "3.10.3"},
        )

        for f in cwd.rglob("*.py"):
            a = f.with_suffix(f".py{ASSERT_SUFFIX}")
            if not a.exists():
                a = f.with_suffix(".py.assert")
            self.assertEqual(f.read_text(), a.read_text(), msg=f.name)
        for a in cwd.rglob("*.py.assert"):
            f = a.with_suffix("")
            self.assertTrue(f.exists(), f"{f} doesn't exist")



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from libc.stdint cimport int64_t, int32_t


DEF MONTHS_PER_YEAR         = 12


@cython.final
cdef class DateDuration:

    def __init__(self, *, int32_t days=0, int32_t months=0):
        self.days = days
        self.months = months

    def __eq__(self, other):
        if type(other) is not DateDuration:
            return NotImplemented

        return (
            self.days == (<DateDuration>other).days and
            self.months == (<DateDuration>other).months
        )

    def __hash__(self):
        return hash((DateDuration, self.days, self.months))

    def __repr__(self):
        return f'<gel.DateDuration "{self}">'

    @cython.cdivision(True)
    def __str__(self):
        # This is closely modeled after interval_out().
        cdef:
            int64_t year = self.months / MONTHS_PER_YEAR
            int64_t mon = self.months % MONTHS_PER_YEAR
            list buf = []

        if not self.months and not self.days:
            return 'P0D'

        buf.append('P')

        if year:
            buf.append(f'{year}Y')

        if mon:
            buf.append(f'{mon}M')

        if self.days:
            buf.append(f'{self.days}D')

        return ''.join(buf)


cdef new_date_duration(int32_t days, int32_t months):
    cdef DateDuration dur = DateDuration.__new__(DateDuration)

    dur.days = days
    dur.months = months

    return dur



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2020-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from __future__ import annotations
import abc
import dataclasses
import typing

from . import datatypes
from . import describe
from . import enums
from . import options
from .protocol import protocol

__all__ = (
    "QueryWithArgs",
    "QueryCache",
    "QueryOptions",
    "QueryContext",
    "Executor",
    "ExecuteContext",
    "AsyncIOExecutor",
    "ReadOnlyExecutor",
    "AsyncIOReadOnlyExecutor",
    "DescribeContext",
    "DescribeResult",
)


class QueryWithArgs(typing.NamedTuple):
    query: str
    args: typing.Tuple
    kwargs: typing.Dict[str, typing.Any]
    input_language: protocol.InputLanguage = protocol.InputLanguage.EDGEQL


class QueryCache(typing.NamedTuple):
    codecs_registry: protocol.CodecsRegistry
    query_cache: protocol.LRUMapping


class QueryOptions(typing.NamedTuple):
    output_format: protocol.OutputFormat
    expect_one: bool
    required_one: bool


class QueryContext(typing.NamedTuple):
    query: QueryWithArgs
    cache: QueryCache
    query_options: QueryOptions
    retry_options: typing.Optional[options.RetryOptions]
    state: typing.Optional[options.State]
    warning_handler: options.WarningHandler
    annotations: typing.Dict[str, str]

    def lower(
        self, *, allow_capabilities: enums.Capability
    ) -> protocol.ExecuteContext:
        return protocol.ExecuteContext(
            query=self.query.query,
            args=self.query.args,
            kwargs=self.query.kwargs,
            reg=self.cache.codecs_registry,
            qc=self.cache.query_cache,
            input_language=self.query.input_language,
            output_format=self.query_options.output_format,
            expect_one=self.query_options.expect_one,
            required_one=self.query_options.required_one,
            allow_capabilities=allow_capabilities,
            state=self.state.as_dict() if self.state else None,
            annotations=self.annotations,
        )


class ExecuteContext(typing.NamedTuple):
    query: QueryWithArgs
    cache: QueryCache
    retry_options: typing.Optional[options.RetryOptions]
    state: typing.Optional[options.State]
    warning_handler: options.WarningHandler
    annotations: typing.Dict[str, str]

    def lower(
        self, *, allow_capabilities: enums.Capability
    ) -> protocol.ExecuteContext:
        return protocol.ExecuteContext(
            query=self.query.query,
            args=self.query.args,
            kwargs=self.query.kwargs,
            reg=self.cache.codecs_registry,
            qc=self.cache.query_cache,
            input_language=self.query.input_language,
            output_format=protocol.OutputFormat.NONE,
            allow_capabilities=allow_capabilities,
            state=self.state.as_dict() if self.state else None,
            annotations=self.annotations,
        )


@dataclasses.dataclass
class DescribeContext:
    query: str
    state: typing.Optional[options.State]
    inject_type_names: bool
    input_language: protocol.InputLanguage
    output_format: protocol.OutputFormat
    expect_one: bool

    def lower(
        self, *, allow_capabilities: enums.Capability
    ) -> protocol.ExecuteContext:
        return protocol.ExecuteContext(
            query=self.query,
            args=None,
            kwargs=None,
            reg=protocol.CodecsRegistry(),
            qc=protocol.LRUMapping(maxsize=1),
            input_language=self.input_language,
            output_format=self.output_format,
            expect_one=self.expect_one,
            inline_typenames=self.inject_type_names,
            allow_capabilities=allow_capabilities,
            state=self.state.as_dict() if self.state else None,
        )


@dataclasses.dataclass
class DescribeResult:
    input_type: typing.Optional[describe.AnyType]
    output_type: typing.Optional[describe.AnyType]
    output_cardinality: enums.Cardinality
    capabilities: enums.Capability


_query_opts = QueryOptions(
    output_format=protocol.OutputFormat.BINARY,
    expect_one=False,
    required_one=False,
)
_query_single_opts = QueryOptions(
    output_format=protocol.OutputFormat.BINARY,
    expect_one=True,
    required_one=False,
)
_query_required_single_opts = QueryOptions(
    output_format=protocol.OutputFormat.BINARY,
    expect_one=True,
    required_one=True,
)
_query_json_opts = QueryOptions(
    output_format=protocol.OutputFormat.JSON,
    expect_one=False,
    required_one=False,
)
_query_single_json_opts = QueryOptions(
    output_format=protocol.OutputFormat.JSON,
    expect_one=True,
    required_one=False,
)
_query_required_single_json_opts = QueryOptions(
    output_format=protocol.OutputFormat.JSON,
    expect_one=True,
    required_one=True,
)


class BaseReadOnlyExecutor(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def _get_query_cache(self) -> QueryCache:
        ...

    @abc.abstractmethod
    def _get_retry_options(self) -> typing.Optional[options.RetryOptions]:
        ...

    @abc.abstractmethod
    def _get_state(self) -> options.State:
        ...

    @abc.abstractmethod
    def _get_warning_handler(self) -> options.WarningHandler:
        ...

    def _get_annotations(self) -> typing.Dict[str, str]:
        return {}


class ReadOnlyExecutor(BaseReadOnlyExecutor):
    """Subclasses can execute *at least* read-only queries"""

    __slots__ = ()

    @abc.abstractmethod
    def _query(self, query_context: QueryContext):
        ...

    def query(self, query: str, *args, **kwargs) -> list:
        return self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    def query_single(
        self, query: str, *args, **kwargs
    ) -> typing.Union[typing.Any, None]:
        return self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_single_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    def query_required_single(self, query: str, *args, **kwargs) -> typing.Any:
        return self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_required_single_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    def query_json(self, query: str, *args, **kwargs) -> str:
        return self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_json_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    def query_single_json(self, query: str, *args, **kwargs) -> str:
        return self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_single_json_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    def query_required_single_json(self, query: str, *args, **kwargs) -> str:
        return self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_required_single_json_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    def query_sql(self, query: str, *args, **kwargs) -> list[datatypes.Record]:
        return self._query(QueryContext(
            query=QueryWithArgs(
                query,
                args,
                kwargs,
                input_language=protocol.InputLanguage.SQL,
            ),
            cache=self._get_query_cache(),
            query_options=_query_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    @abc.abstractmethod
    def _execute(self, execute_context: ExecuteContext):
        ...

    def execute(self, commands: str, *args, **kwargs) -> None:
        self._execute(ExecuteContext(
            query=QueryWithArgs(commands, args, kwargs),
            cache=self._get_query_cache(),
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    def execute_sql(self, commands: str, *args, **kwargs) -> None:
        self._execute(ExecuteContext(
            query=QueryWithArgs(
                commands,
                args,
                kwargs,
                input_language=protocol.InputLanguage.SQL,
            ),
            cache=self._get_query_cache(),
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))


class Executor(ReadOnlyExecutor):
    """Subclasses can execute both read-only and modification queries"""

    __slots__ = ()


class AsyncIOReadOnlyExecutor(BaseReadOnlyExecutor):
    """Subclasses can execute *at least* read-only queries"""

    __slots__ = ()

    @abc.abstractmethod
    async def _query(self, query_context: QueryContext):
        ...

    async def query(self, query: str, *args, **kwargs) -> list:
        return await self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    async def query_single(self, query: str, *args, **kwargs) -> typing.Any:
        return await self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_single_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    async def query_required_single(
        self,
        query: str,
        *args,
        **kwargs
    ) -> typing.Any:
        return await self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_required_single_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    async def query_json(self, query: str, *args, **kwargs) -> str:
        return await self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_json_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    async def query_single_json(self, query: str, *args, **kwargs) -> str:
        return await self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_single_json_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    async def query_required_single_json(
        self,
        query: str,
        *args,
        **kwargs
    ) -> str:
        return await self._query(QueryContext(
            query=QueryWithArgs(query, args, kwargs),
            cache=self._get_query_cache(),
            query_options=_query_required_single_json_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    async def query_sql(self, query: str, *args, **kwargs) -> typing.Any:
        return await self._query(QueryContext(
            query=QueryWithArgs(
                query,
                args,
                kwargs,
                input_language=protocol.InputLanguage.SQL,
            ),
            cache=self._get_query_cache(),
            query_options=_query_opts,
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    @abc.abstractmethod
    async def _execute(self, execute_context: ExecuteContext) -> None:
        ...

    async def execute(self, commands: str, *args, **kwargs) -> None:
        await self._execute(ExecuteContext(
            query=QueryWithArgs(commands, args, kwargs),
            cache=self._get_query_cache(),
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))

    async def execute_sql(self, commands: str, *args, **kwargs) -> None:
        await self._execute(ExecuteContext(
            query=QueryWithArgs(
                commands,
                args,
                kwargs,
                input_language=protocol.InputLanguage.SQL,
            ),
            cache=self._get_query_cache(),
            retry_options=self._get_retry_options(),
            state=self._get_state(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))


class AsyncIOExecutor(AsyncIOReadOnlyExecutor):
    """Subclasses can execute both read-only and modification queries"""

    __slots__ = ()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2022-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import getpass
import io
import os
import pathlib
import sys
import textwrap
import typing

import gel
from gel import abstract
from gel import describe
from gel.con_utils import find_gel_project_dir
from gel.color import get_color


C = get_color()
SYS_VERSION_INFO = os.getenv("EDGEDB_PYTHON_CODEGEN_PY_VER")
if SYS_VERSION_INFO:
    SYS_VERSION_INFO = tuple(map(int, SYS_VERSION_INFO.split(".")))[:2]
else:
    SYS_VERSION_INFO = sys.version_info[:2]

INDENT = "    "
SUFFIXES = [
    ("async", "_async_edgeql.py", True),
    ("blocking", "_edgeql.py", False),
]
FILE_MODE_OUTPUT_FILE = "generated"

TYPE_MAPPING = {
    "std::str": "str",
    "std::float32": "float",
    "std::float64": "float",
    "std::int16": "int",
    "std::int32": "int",
    "std::int64": "int",
    "std::bigint": "int",
    "std::bool": "bool",
    "std::uuid": "uuid.UUID",
    "std::bytes": "bytes",
    "std::decimal": "decimal.Decimal",
    "std::datetime": "datetime.datetime",
    "std::duration": "datetime.timedelta",
    "std::json": "str",
    "cal::local_date": "datetime.date",
    "cal::local_time": "datetime.time",
    "cal::local_datetime": "datetime.datetime",
    "cal::relative_duration": "gel.RelativeDuration",
    "cal::date_duration": "gel.DateDuration",
    "cfg::memory": "gel.ConfigMemory",
    "ext::pgvector::vector": "array.array",
}

TYPE_IMPORTS = {
    "std::uuid": "uuid",
    "std::decimal": "decimal",
    "std::datetime": "datetime",
    "std::duration": "datetime",
    "cal::local_date": "datetime",
    "cal::local_time": "datetime",
    "cal::local_datetime": "datetime",
    "ext::pgvector::vector": "array",
}

INPUT_TYPE_MAPPING = TYPE_MAPPING.copy()
INPUT_TYPE_MAPPING.update(
    {
        "ext::pgvector::vector": "typing.Sequence[float]",
    }
)

INPUT_TYPE_IMPORTS = TYPE_IMPORTS.copy()
INPUT_TYPE_IMPORTS.update(
    {
        "ext::pgvector::vector": "typing",
    }
)

PYDANTIC_MIXIN = """\
class NoPydanticValidation:
    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type, _handler):
        # Pydantic 2.x
        from pydantic_core.core_schema import any_schema
        return any_schema()

    @classmethod
    def __get_validators__(cls):
        # Pydantic 1.x
        from pydantic.dataclasses import dataclass as pydantic_dataclass
        _ = pydantic_dataclass(cls)
        cls.__pydantic_model__.__get_validators__ = lambda: []
        return []\
"""


def print_msg(msg):
    print(msg, file=sys.stderr)


def print_error(msg):
    print_msg(f"{C.BOLD}{C.FAIL}error: {C.ENDC}{C.BOLD}{msg}{C.ENDC}")


def _get_conn_args(args: argparse.Namespace):
    if args.password_from_stdin:
        if args.password:
            print_error(
                "--password and --password-from-stdin are "
                "mutually exclusive",
            )
            sys.exit(22)
        if sys.stdin.isatty():
            password = getpass.getpass()
        else:
            password = sys.stdin.read().strip()
    else:
        password = args.password
    if args.dsn and args.instance:
        print_error("--dsn and --instance are mutually exclusive")
        sys.exit(22)
    return dict(
        dsn=args.dsn or args.instance,
        credentials_file=args.credentials_file,
        host=args.host,
        port=args.port,
        database=args.database,
        user=args.user,
        password=password,
        tls_ca_file=args.tls_ca_file,
        tls_security=args.tls_security,
    )


class Generator:
    def __init__(self, args: argparse.Namespace):
        self._default_module = "default"
        self._targets = args.target
        self._skip_pydantic_validation = args.skip_pydantic_validation
        self._async = False
        try:
            self._project_dir = pathlib.Path(find_gel_project_dir())
        except gel.ClientConnectionError:
            print(
                "Cannot find gel.toml: "
                "codegen must be run under an EdgeDB project dir"
            )
            sys.exit(2)
        print_msg(f"Found EdgeDB project: {C.BOLD}{self._project_dir}{C.ENDC}")
        self._client = gel.create_client(**_get_conn_args(args))
        self._single_mode_files = args.file
        self._search_dirs = []
        for search_dir in args.dir or []:
            search_dir = pathlib.Path(search_dir).absolute()
            if (
                search_dir == self._project_dir
                or self._project_dir in search_dir.parents
            ):
                self._search_dirs.append(search_dir)
            else:
                print(
                    f"--dir '{search_dir}' is not under "
                    f"the project directory: {self._project_dir}"
                )
                sys.exit(1)
        self._method_names = set()
        self._describe_results = []

        self._cache = {}
        self._imports = set()
        self._aliases = {}
        self._defs = {}
        self._names = set()
        self._use_pydantic = False

    def _new_file(self):
        self._cache.clear()
        self._imports.clear()
        self._aliases.clear()
        self._defs.clear()
        self._names.clear()
        self._use_pydantic = False

    def run(self):
        try:
            self._client.ensure_connected()
        except gel.EdgeDBError as e:
            print(f"Failed to connect to EdgeDB instance: {e}")
            sys.exit(61)
        with self._client:
            if self._search_dirs:
                for search_dir in self._search_dirs:
                    self._process_dir(search_dir)
            else:
                self._process_dir(self._project_dir)
        for target, suffix, is_async in SUFFIXES:
            if target in self._targets:
                self._async = is_async
                if self._single_mode_files:
                    self._generate_single_file(suffix)
                else:
                    self._generate_files(suffix)
                self._new_file()
        print_msg(f"{C.GREEN}{C.BOLD}Done.{C.ENDC}")

    def _process_dir(self, dir_: pathlib.Path):
        for file_or_dir in dir_.iterdir():
            if not file_or_dir.exists():
                continue
            if file_or_dir.is_dir():
                if file_or_dir.relative_to(self._project_dir) != pathlib.Path(
                    "dbschema/migrations"
                ):
                    self._process_dir(file_or_dir)
            elif file_or_dir.suffix.lower() == ".edgeql":
                self._process_file(file_or_dir)

    def _process_file(self, source: pathlib.Path):
        print_msg(f"{C.BOLD}Processing{C.ENDC} {C.BLUE}{source}{C.ENDC}")
        with source.open() as f:
            query = f.read()
        name = source.stem
        if self._single_mode_files:
            if name in self._method_names:
                print_error(f"Conflict method names: {name}")
                sys.exit(17)
            self._method_names.add(name)
        dr = self._client._describe_query(query, inject_type_names=True)
        self._describe_results.append((name, source, query, dr))

    def _generate_files(self, suffix: str):
        for name, source, query, dr in self._describe_results:
            target = source.parent / f"{name}{suffix}"
            print_msg(f"{C.BOLD}Generating{C.ENDC} {C.BLUE}{target}{C.ENDC}")
            self._new_file()
            content = self._generate(name, query, dr)
            buf = io.StringIO()
            self._write_comments(buf, [source])
            self._write_definitions(buf)
            buf.write(content)
            with target.open("w") as f:
                f.write(buf.getvalue())

    def _generate_single_file(self, suffix: str):
        print_msg(f"{C.BOLD}Generating single file output...{C.ENDC}")
        buf = io.StringIO()
        output = []
        sources = []
        for name, source, query, dr in sorted(self._describe_results):
            sources.append(source)
            output.append(self._generate(name, query, dr))
        self._write_comments(buf, sources)
        self._write_definitions(buf)
        for i, o in enumerate(output):
            buf.write(o)
            if i < len(output) - 1:
                print(file=buf)
                print(file=buf)

        for target in self._single_mode_files:
            if target:
                target = pathlib.Path(target).absolute()
            else:
                target = self._project_dir / f"{FILE_MODE_OUTPUT_FILE}{suffix}"
            print_msg(f"{C.BOLD}Writing{C.ENDC} {C.BLUE}{target}{C.ENDC}")
            with target.open("w") as f:
                f.write(buf.getvalue())

    def _write_comments(
        self, f: io.TextIOBase, src: typing.List[pathlib.Path]
    ):
        src_str = map(
            lambda p: repr(p.relative_to(self._project_dir).as_posix()), src
        )
        if len(src) > 1:
            print("# AUTOGENERATED FROM:", file=f)
            for s in src_str:
                print(f"#     {s}", file=f)
            print("# WITH:", file=f)
        else:
            print(f"# AUTOGENERATED FROM {next(src_str)} WITH:", file=f)
        cmd = []
        if sys.argv[0].endswith("__main__.py"):
            cmd.append(pathlib.Path(sys.executable).name)
            cmd.extend(["-m", "gel.codegen"])
        else:
            cmd.append(pathlib.Path(sys.argv[0]).name)
        cmd.extend(sys.argv[1:])
        cmd = " ".join(cmd)
        print(f"#     $ {cmd}", file=f)
        print(file=f)
        print(file=f)

    def _write_definitions(self, f: io.TextIOBase):
        print("from __future__ import annotations", file=f)
        for m in sorted(self._imports):
            print(f"import {m}", file=f)
        print(file=f)
        print(file=f)

        if self._aliases:
            for _, a in sorted(self._aliases.items()):
                print(a, file=f)
            print(file=f)
            print(file=f)

        if self._use_pydantic:
            print(PYDANTIC_MIXIN, file=f)
            print(file=f)
            print(file=f)

        for _, d in sorted(self._defs.items()):
            print(d, file=f)
            print(file=f)
            print(file=f)

    def _generate(
        self, name: str, query: str, dr: abstract.DescribeResult
    ) -> str:
        buf = io.StringIO()

        name_hint = f"{self._snake_to_camel(name)}Result"
        out_type = self._generate_code(dr.output_type, name_hint)
        if dr.output_cardinality.is_multi():
            if SYS_VERSION_INFO >= (3, 9):
                out_type = f"list[{out_type}]"
            else:
                self._imports.add("typing")
                out_type = f"typing.List[{out_type}]"
        elif dr.output_cardinality == gel.Cardinality.AT_MOST_ONE:
            if SYS_VERSION_INFO >= (3, 10):
                out_type = f"{out_type} | None"
            else:
                self._imports.add("typing")
                out_type = f"typing.Optional[{out_type}]"

        args = {}
        kw_only = False
        if isinstance(dr.input_type, describe.ObjectType):
            if "".join(dr.input_type.elements.keys()).isdecimal():
                for el_name, el in dr.input_type.elements.items():
                    args[int(el_name)] = self._generate_code_with_cardinality(
                        el.type, f"arg{el_name}", el.cardinality, is_input=True
                    )
                args = {f"arg{i}": v for i, v in sorted(args.items())}
            else:
                kw_only = True
                for el_name, el in dr.input_type.elements.items():
                    args[el_name] = self._generate_code_with_cardinality(
                        el.type,
                        el_name,
                        el.cardinality,
                        keyword_argument=True,
                        is_input=True
                    )

        if self._async:
            print(f"async def {name}(", file=buf)
        else:
            print(f"def {name}(", file=buf)
        self._imports.add("gel")
        if self._async:
            print(f"{INDENT}executor: gel.AsyncIOExecutor,", file=buf)
        else:
            print(f"{INDENT}executor: gel.Executor,", file=buf)
        if kw_only:
            print(f"{INDENT}*,", file=buf)
        for name, arg in args.items():
            print(f"{INDENT}{name}: {arg},", file=buf)
        print(f") -> {out_type}:", file=buf)
        if dr.output_cardinality.is_multi():
            method = "query"
            rt = "return "
        elif dr.output_cardinality == gel.Cardinality.NO_RESULT:
            method = "execute"
            rt = ""
        else:
            method = "query_single"
            rt = "return "

        if self._async:
            print(f"{INDENT}{rt}await executor.{method}(", file=buf)
        else:
            print(f"{INDENT}{rt}executor.{method}(", file=buf)
        print(f'{INDENT}{INDENT}"""\\', file=buf)
        print(
            textwrap.indent(
                textwrap.dedent(query).strip(), f"{INDENT}{INDENT}"
            )
            + "\\",
            file=buf,
        )
        print(f'{INDENT}{INDENT}""",', file=buf)
        for name in args:
            if kw_only:
                print(f"{INDENT}{INDENT}{name}={name},", file=buf)
            else:
                print(f"{INDENT}{INDENT}{name},", file=buf)
        print(f"{INDENT})", file=buf)
        return buf.getvalue()

    def _generate_code(
        self,
        type_: typing.Optional[describe.AnyType],
        name_hint: str,
        is_input: bool = False,
    ) -> str:
        if type_ is None:
            return "None"

        if (type_.desc_id, is_input) in self._cache:
            return self._cache[(type_.desc_id, is_input)]

        imports = INPUT_TYPE_IMPORTS if is_input else TYPE_IMPORTS
        mapping = INPUT_TYPE_MAPPING if is_input else TYPE_MAPPING

        if isinstance(type_, describe.BaseScalarType):
            if import_str := imports.get(type_.name):
                self._imports.add(import_str)
            rv = mapping[type_.name]

        elif isinstance(type_, describe.SequenceType):
            el_type = self._generate_code(
                type_.element_type, f"{name_hint}Item", is_input
            )
            if SYS_VERSION_INFO >= (3, 9):
                rv = f"list[{el_type}]"
            else:
                self._imports.add("typing")
                rv = f"typing.List[{el_type}]"

        elif isinstance(type_, describe.TupleType):
            elements = ", ".join(
                self._generate_code(el_type, f"{name_hint}Item", is_input)
                for el_type in type_.element_types
            )
            if SYS_VERSION_INFO >= (3, 9):
                rv = f"tuple[{elements}]"
            else:
                self._imports.add("typing")
                rv = f"typing.Tuple[{elements}]"

        elif isinstance(type_, describe.ScalarType):
            rv = self._find_name(type_.name or name_hint)
            base_type_name = type_.base_type.name
            if import_str := imports.get(base_type_name):
                self._imports.add(import_str)
            value = mapping[base_type_name]
            self._aliases[rv] = f"{rv} = {value}"

        elif isinstance(type_, describe.ObjectType):
            rv = self._find_name(name_hint)
            buf = io.StringIO()
            self._imports.add("dataclasses")
            print("@dataclasses.dataclass", file=buf)
            if self._skip_pydantic_validation:
                print(f"class {rv}(NoPydanticValidation):", file=buf)
                self._use_pydantic = True
            else:
                print(f"class {rv}:", file=buf)
            link_props = []
            for el_name, element in type_.elements.items():
                if element.is_implicit and el_name != "id":
                    continue
                name_hint = f"{rv}{self._snake_to_camel(el_name)}"
                el_code = self._generate_code_with_cardinality(
                    element.type, name_hint, element.cardinality
                )
                if element.kind == gel.ElementKind.LINK_PROPERTY:
                    link_props.append((el_name, el_code))
                else:
                    print(f"{INDENT}{el_name}: {el_code}", file=buf)
            if link_props:
                print(file=buf)
                self._imports.add("typing")
                typing_literal = "typing.Literal"
                for el_name, el_code in link_props:
                    print(f"{INDENT}@typing.overload", file=buf)
                    print(
                        f'{INDENT}def __getitem__'
                        f'(self, key: {typing_literal}["{el_name}"]) '
                        f'-> {el_code}:',
                        file=buf,
                    )
                    print(f"{INDENT}{INDENT}...", file=buf)
                    print(file=buf)
                print(
                    f"{INDENT}def __getitem__(self, key: str) -> typing.Any:",
                    file=buf,
                )
                print(
                    f"{INDENT}{INDENT}raise NotImplementedError", file=buf
                )

            self._defs[rv] = buf.getvalue().strip()

        elif isinstance(type_, describe.NamedTupleType):
            rv = self._find_name(name_hint)
            buf = io.StringIO()
            self._imports.add("typing")
            print(f"class {rv}(typing.NamedTuple):", file=buf)
            for el_name, el_type in type_.element_types.items():
                el_code = self._generate_code(
                    el_type, f"{rv}{self._snake_to_camel(el_name)}", is_input
                )
                print(f"{INDENT}{el_name}: {el_code}", file=buf)
            self._defs[rv] = buf.getvalue().strip()

        elif isinstance(type_, describe.EnumType):
            rv = self._find_name(type_.name or name_hint)
            buf = io.StringIO()
            self._imports.add("enum")
            print(f"class {rv}(enum.Enum):", file=buf)
            for member, member_id in self._to_unique_idents(type_.members):
                print(f'{INDENT}{member_id.upper()} = "{member}"', file=buf)
            self._defs[rv] = buf.getvalue().strip()

        elif isinstance(type_, describe.RangeType):
            value = self._generate_code(type_.value_type, name_hint, is_input)
            rv = f"gel.Range[{value}]"

        else:
            rv = "??"

        self._cache[(type_.desc_id, is_input)] = rv
        return rv

    def _generate_code_with_cardinality(
        self,
        type_: typing.Optional[describe.AnyType],
        name_hint: str,
        cardinality: gel.Cardinality,
        keyword_argument: bool = False,
        is_input: bool = False,
    ):
        rv = self._generate_code(type_, name_hint, is_input)
        if cardinality == gel.Cardinality.AT_MOST_ONE:
            if SYS_VERSION_INFO >= (3, 10):
                rv = f"{rv} | None"
            else:
                self._imports.add("typing")
                rv = f"typing.Optional[{rv}]"
            if keyword_argument:
                rv = f"{rv} = None"
        return rv

    def _find_name(self, name: str) -> str:
        default_prefix = f"{self._default_module}::"
        if name.startswith(default_prefix):
            name = name[len(default_prefix) :]
        mod, _, name = name.rpartition("::")
        name = self._snake_to_camel(name)
        name = mod.title() + name
        if name in self._names:
            for i in range(2, 100):
                new = f"{name}{i:02d}"
                if new not in self._names:
                    name = new
                    break
            else:
                print_error(f"Failed to find a unique name for: {name}")
                sys.exit(17)
        self._names.add(name)
        return name

    def _snake_to_camel(self, name: str) -> str:
        parts = name.split("_")
        if len(parts) > 1 or name.islower():
            return "".join(map(str.title, parts))
        else:
            return name

    def _to_unique_idents(
        self, names: typing.Iterable[typing.Tuple[str, str]]
    ) -> typing.Iterator[str]:
        dedup = set()
        for name in names:
            if name.isidentifier():
                name_id = name
                sep = name.endswith("_")
            else:
                sep = True
                result = []
                for i, c in enumerate(name):
                    if c.isdigit():
                        if i == 0:
                            result.append("e_")
                        result.append(c)
                        sep = False
                    elif c.isidentifier():
                        result.append(c)
                        sep = c == "_"
                    elif not sep:
                        result.append("_")
                        sep = True
                name_id = "".join(result)
            rv = name_id
            if not sep:
                name_id = name_id + "_"
            i = 1
            while rv in dedup:
                rv = f"{name_id}{i}"
                i += 1
            dedup.add(rv)
            yield name, rv



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import socket
import time

from gel.pgproto.pgproto cimport (
    WriteBuffer,
    ReadBuffer,
)

from .. import con_utils
from .. import errors
from . cimport protocol


DEF RECV_BUF = 65536


cdef class BlockingIOProtocol(protocol.SansIOProtocolBackwardsCompatible):

    def __init__(self, con_params, sock):
        protocol.SansIOProtocolBackwardsCompatible.__init__(self, con_params)
        self.sock = sock
        self.deadline = 0

    cpdef abort(self):
        self.terminate()
        self._disconnect()

    cdef _disconnect(self):
        self.connected = False
        sock, self.sock = self.sock, None
        if sock is not None:
            try:
                sock.shutdown(socket.SHUT_RDWR)
            except OSError:
                pass
            sock.close()

    cdef write(self, WriteBuffer buf):
        try:
            self.sock.send(buf)
        except OSError as e:
            self._disconnect()
            raise con_utils.wrap_error(e) from e

    async def wait_for_message(self):
        cdef float timeout
        if self.deadline > 0:
            while not self.buffer.take_message():
                timeout = self.deadline - time.monotonic()
                if timeout <= 0:
                    self.abort()
                    raise TimeoutError
                try:
                    self.sock.settimeout(timeout)
                    data = self.sock.recv(RECV_BUF)
                except OSError as e:
                    self._disconnect()
                    raise con_utils.wrap_error(e) from e
                if not data:
                    self._disconnect()
                    raise errors.ClientConnectionClosedError()
                self.buffer.feed_data(data)
        else:
            while not self.buffer.take_message():
                try:
                    data = self.sock.recv(RECV_BUF)
                except OSError as e:
                    self._disconnect()
                    raise con_utils.wrap_error(e) from e
                if not data:
                    self._disconnect()
                    raise errors.ClientConnectionClosedError()
                self.buffer.feed_data(data)
        self.last_active_timestamp = time.monotonic()

    async def try_recv_eagerly(self):
        if self.buffer.take_message():
            return

        self.sock.settimeout(0)  # Make non-blocking.
        try:
            while not self.buffer.take_message():
                data = self.sock.recv(RECV_BUF)
                if not data:
                    self._disconnect()
                    raise errors.ClientConnectionClosedError()
                self.buffer.feed_data(data)
        except BlockingIOError:
            # No data in the socket net buffer.
            return
        except OSError as e:
            self._disconnect()
            raise con_utils.wrap_error(e) from e
        else:
            self.last_active_timestamp = time.monotonic()
        finally:
            self.sock.settimeout(None)

    async def wait_for_connect(self):
        return True

    async def wait_for_disconnect(self):
        if self.cancelled or not self.connected:
            return
        try:
            while True:
                if not self.buffer.take_message():
                    await self.wait_for_message()
                self.fallthrough()
        except errors.ClientConnectionClosedError:
            pass

    async def wait_for(self, coro, timeout):
        if timeout is None:
            return await coro
        else:
            self.deadline = time.monotonic() + timeout
            try:
                return await coro
            finally:
                self.deadline = 0
                if self.sock is not None:
                    self.sock.settimeout(None)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from gel import _testbase as tb
from edgedb import errors


class TestGlobals(tb.AsyncQueryTestCase):

    async def test_globals_01(self):
        db = self.client
        if db.is_proto_lt_1_0:
            self.skipTest("Global is added in EdgeDB 2.0")

        await db.execute('''
            CREATE GLOBAL glob -> str;
            CREATE REQUIRED GLOBAL req_glob -> str {
                SET default := '!';
            };
            CREATE GLOBAL def_glob -> str {
                SET default := '!';
            };
            CREATE GLOBAL computed := '!';
            CREATE MODULE foo;
            CREATE MODULE foo::bar;
            CREATE GLOBAL foo::bar::baz -> str;
        ''')

        async with db.with_globals(glob='test') as gdb:
            x = await gdb.query_single('select global glob')
            self.assertEqual(x, 'test')

            x = await gdb.query_single('select global req_glob')
            self.assertEqual(x, '!')

            x = await gdb.query_single('select global def_glob')
            self.assertEqual(x, '!')

        async with db.with_globals(req_glob='test') as gdb:
            x = await gdb.query_single('select global req_glob')
            self.assertEqual(x, 'test')

        async with db.with_globals(def_glob='test') as gdb:
            x = await gdb.query_single('select global def_glob')
            self.assertEqual(x, 'test')

        # Setting def_glob explicitly to None should override
        async with db.with_globals(def_glob=None) as gdb:
            x = await gdb.query_single('select global def_glob')
            self.assertEqual(x, None)

        # Setting computed global should produce error
        async with db.with_globals(computed='test') as gdb:
            with self.assertRaises(errors.QueryArgumentError):
                await gdb.query_single('select global computed')

        async with db.with_globals({'foo::bar::baz': 'asdf'}) as gdb:
            x = await gdb.query_single('select global foo::bar::baz')
            self.assertEqual(x, 'asdf')

    async def test_client_state_mismatch(self):
        db = self.client
        if db.is_proto_lt_1_0:
            self.skipTest("State over protocol is added in EdgeDB 2.0")

        await db.execute('create global mglob -> int32')

        c = self.make_test_client(database=self.get_database_name())
        c = c.with_globals(mglob=42)
        self.assertEqual(await c.query_single('select global mglob'), 42)

        await db.execute('create global mglob2 -> str')
        self.assertEqual(await c.query_single('select global mglob'), 42)

        await db.execute('alter global mglob set type str reset to default')
        with self.assertRaises(errors.InvalidArgumentError):
            await c.query_single('select global mglob')

        await c.aclose()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2024-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import typing

import dataclasses as dc
import enum


class ChatParticipantRole(enum.Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    TOOL = "tool"


class Custom(typing.TypedDict):
    role: ChatParticipantRole
    content: str


class Prompt:
    name: typing.Optional[str]
    id: typing.Optional[str]
    custom: typing.Optional[typing.List[Custom]]


@dc.dataclass
class RAGOptions:
    model: str
    prompt: typing.Optional[Prompt] = None

    def derive(self, kwargs):
        return RAGOptions(**{**dc.asdict(self), **kwargs})


@dc.dataclass
class QueryContext:
    query: str = ""
    variables: typing.Optional[typing.Dict[str, typing.Any]] = None
    globals: typing.Optional[typing.Dict[str, typing.Any]] = None
    max_object_count: typing.Optional[int] = None

    def derive(self, kwargs):
        return QueryContext(**{**dc.asdict(self), **kwargs})


@dc.dataclass
class RAGRequest:
    model: str
    prompt: typing.Optional[Prompt]
    context: QueryContext
    query: str
    stream: typing.Optional[bool]

    def to_httpx_request(self) -> typing.Dict[str, typing.Any]:
        return dict(
            url="/rag",
            headers={
                "Content-Type": "application/json",
                "Accept": (
                    "text/event-stream" if self.stream else "application/json"
                ),
            },
            json=dc.asdict(self),
        )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2024-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime as dt
import os
import uuid
import unittest

try:
    from sqlalchemy import create_engine, select
    from sqlalchemy.orm import Session
except ImportError:
    NO_ORM = True
else:
    NO_ORM = False

from gel import _testbase as tb
from gel.orm import sqla


class TestSQLABasic(tb.SQLATestCase):
    SCHEMA = os.path.join(os.path.dirname(__file__), 'dbsetup',
                          'base.esdl')

    SETUP = os.path.join(os.path.dirname(__file__), 'dbsetup',
                         'base.edgeql')

    MODEL_PACKAGE = 'basemodels'

    @classmethod
    def setUpClass(cls):
        if NO_ORM:
            raise unittest.SkipTest("sqlalchemy is not installed")

        super().setUpClass()
        cls.engine = create_engine(cls.get_dsn_for_sqla())
        cls.sess = Session(cls.engine, autobegin=False)

        from basemodels import default
        cls.sm = default

    def setUp(self):
        super().setUp()

        if self.client.query_required_single('''
            select sys::get_version().major < 6
        '''):
            self.skipTest("Test needs SQL DML queries")

        self.sess.begin()

    def tearDown(self):
        super().tearDown()
        self.sess.rollback()

    def test_sqla_read_models_01(self):
        vals = {r.name for r in self.sess.query(self.sm.User).all()}
        self.assertEqual(
            vals, {'Alice', 'Billie', 'Cameron', 'Dana', 'Elsa', 'Zoe'})

        vals = {r.name for r in self.sess.query(self.sm.UserGroup).all()}
        self.assertEqual(
            vals, {'red', 'green', 'blue'})

        vals = {r.num for r in self.sess.query(self.sm.GameSession).all()}
        self.assertEqual(vals, {123, 456})

        vals = {r.body for r in self.sess.query(self.sm.Post).all()}
        self.assertEqual(
            vals, {'Hello', "I'm Alice", "I'm Cameron", '*magic stuff*'})

        # Read from the abstract type
        vals = {r.name for r in self.sess.query(self.sm.Named).all()}
        self.assertEqual(
            vals,
            {
                'Alice', 'Billie', 'Cameron', 'Dana', 'Elsa', 'Zoe',
                'red', 'green', 'blue',
            }
        )

    def test_sqla_read_models_02(self):
        # test single link and the one-to-many backlink
        # using load-on-demand

        res = self.sess.query(self.sm.Post).all()
        vals = {(p.author.name, p.body) for p in res}
        self.assertEqual(
            vals,
            {
                ('Alice', 'Hello'),
                ('Alice', "I'm Alice"),
                ('Cameron', "I'm Cameron"),
                ('Elsa', '*magic stuff*'),
            }
        )

        # use backlink
        res = self.sess.query(self.sm.User).order_by(self.sm.User.name).all()
        vals = [
            (u.name, {p.body for p in u._author_Post})
            for u in res
        ]
        self.assertEqual(
            vals,
            [
                ('Alice', {'Hello', "I'm Alice"}),
                ('Billie', set()),
                ('Cameron', {"I'm Cameron"}),
                ('Dana', set()),
                ('Elsa', {'*magic stuff*'}),
                ('Zoe', set()),
            ]
        )

    def test_sqla_read_models_03(self):
        # test single link and the one-to-many backlink

        res = self.sess.execute(
            select(self.sm.Post, self.sm.User)
            .join(self.sm.Post.author)
        )
        vals = {(p.author.name, p.body) for (p, _) in res}
        self.assertEqual(
            vals,
            {
                ('Alice', 'Hello'),
                ('Alice', "I'm Alice"),
                ('Cameron', "I'm Cameron"),
                ('Elsa', '*magic stuff*'),
            }
        )

        # join via backlink
        res = self.sess.execute(
            select(self.sm.Post, self.sm.User)
            .join(self.sm.User._author_Post)
            .order_by(self.sm.Post.body)
        )
        # We'll get a cross-product, so we need to jump through some hoops to
        # remove duplicates
        vals = {
            (u.name, tuple(p.body for p in u._author_Post))
            for (_, u) in res
        }
        self.assertEqual(
            vals,
            {
                ('Alice', ('Hello', "I'm Alice")),
                ('Cameron', ("I'm Cameron",)),
                ('Elsa', ('*magic stuff*',)),
            }
        )

        # LEFT OUTER join via backlink
        res = self.sess.execute(
            select(self.sm.Post, self.sm.User)
            .join(self.sm.User._author_Post, isouter=True)
            .order_by(self.sm.Post.body)
        )
        vals = {
            (u.name, tuple(p.body for p in u._author_Post))
            for (p, u) in res
        }
        self.assertEqual(
            vals,
            {
                ('Alice', ('Hello', "I'm Alice")),
                ('Billie', ()),
                ('Cameron', ("I'm Cameron",)),
                ('Dana', ()),
                ('Elsa', ('*magic stuff*',)),
                ('Zoe', ()),
            }
        )

    def test_sqla_read_models_04(self):
        # test exclusive multi link and its backlink
        # using load-on-demand

        res = self.sess.query(
            self.sm.GameSession
        ).order_by(self.sm.GameSession.num).all()

        vals = [(g.num, {u.name for u in g.players}) for g in res]
        self.assertEqual(
            vals,
            [
                (123, {'Alice', 'Billie'}),
                (456, {'Dana'}),
            ]
        )

        # use backlink
        res = self.sess.query(self.sm.User).all()
        vals = {
            (u.name, tuple(g.num for g in u._players_GameSession))
            for u in res
        }
        self.assertEqual(
            vals,
            {
                ('Alice', (123,)),
                ('Billie', (123,)),
                ('Cameron', ()),
                ('Dana', (456,)),
                ('Elsa', ()),
                ('Zoe', ()),
            }
        )

    def test_sqla_read_models_05(self):
        # test exclusive multi link and its backlink

        res = self.sess.execute(
            select(self.sm.GameSession, self.sm.User)
            .join(self.sm.GameSession.players)
        )
        # We'll get a cross-product, so we need to jump through some hoops to
        # remove duplicates
        vals = {
            (g.num, tuple(sorted(u.name for u in g.players)))
            for (g, _) in res
        }
        self.assertEqual(
            vals,
            {
                (123, ('Alice', 'Billie')),
                (456, ('Dana',)),
            }
        )

        # LEFT OUTER join via backlink
        res = self.sess.execute(
            select(self.sm.GameSession, self.sm.User)
            .join(self.sm.User._players_GameSession, isouter=True)
        )
        vals = {
            (u.name, tuple(g.num for g in u._players_GameSession))
            for (_, u) in res
        }
        self.assertEqual(
            vals,
            {
                ('Alice', (123,)),
                ('Billie', (123,)),
                ('Cameron', ()),
                ('Dana', (456,)),
                ('Elsa', ()),
                ('Zoe', ()),
            }
        )

    def test_sqla_read_models_06(self):
        # test multi link and its backlink
        # using load-on-demand

        res = self.sess.query(
            self.sm.UserGroup
        ).order_by(self.sm.UserGroup.name).all()

        vals = [(g.name, {u.name for u in g.users}) for g in res]
        self.assertEqual(
            vals,
            [
                ('blue', set()),
                ('green', {'Alice', 'Billie'}),
                ('red', {'Alice', 'Billie', 'Cameron', 'Dana'}),
            ]
        )

        # use backlink
        res = self.sess.query(self.sm.User).order_by(self.sm.User.name).all()
        vals = [
            (u.name, {g.name for g in u._users_UserGroup})
            for u in res
        ]
        self.assertEqual(
            vals,
            [
                ('Alice', {'red', 'green'}),
                ('Billie', {'red', 'green'}),
                ('Cameron', {'red'}),
                ('Dana', {'red'}),
                ('Elsa', set()),
                ('Zoe', set()),
            ]
        )

    def test_sqla_read_models_07(self):
        # test exclusive multi link and its backlink

        res = self.sess.execute(
            select(self.sm.UserGroup, self.sm.User)
            .join(self.sm.UserGroup.users, isouter=True)
            .order_by(self.sm.UserGroup.name)
        )
        # We'll get a cross-product, so we need to jump through some hoops to
        # remove duplicates
        vals = {
            (g.name, tuple(sorted(u.name for u in g.users)))
            for (g, _) in res
        }
        self.assertEqual(
            vals,
            {
                ('blue', ()),
                ('green', ('Alice', 'Billie')),
                ('red', ('Alice', 'Billie', 'Cameron', 'Dana')),
            }
        )

        # LEFT OUTER join via backlink
        res = self.sess.execute(
            select(self.sm.UserGroup, self.sm.User)
            .join(self.sm.User._users_UserGroup, isouter=True)
        )
        vals = {
            (u.name, tuple(sorted(g.name for g in u._users_UserGroup)))
            for (_, u) in res
        }
        self.assertEqual(
            vals,
            {
                ('Alice', ('green', 'red')),
                ('Billie', ('green', 'red')),
                ('Cameron', ('red',)),
                ('Dana', ('red',)),
                ('Elsa', ()),
                ('Zoe', ()),
            }
        )

    def test_sqla_read_models_08(self):
        # test arrays, bytes and various date/time scalars

        res = self.sess.query(self.sm.AssortedScalars).one()

        self.assertEqual(res.name, 'hello world')
        self.assertEqual(res.vals, ['brown', 'fox'])
        self.assertEqual(res.bstr, b'word\x00\x0b')
        self.assertEqual(
            res.time,
            dt.time(20, 13, 45, 678_000),
        )
        self.assertEqual(
            res.date,
            dt.date(2025, 1, 26),
        )
        # time zone aware
        self.assertEqual(
            res.ts,
            dt.datetime.fromisoformat('2025-01-26T20:13:45+00:00'),
        )
        # naive datetime
        self.assertEqual(
            res.lts,
            dt.datetime.fromisoformat('2025-01-26T20:13:45'),
        )

    def test_sqla_create_models_01(self):
        vals = self.sess.query(self.sm.User).filter_by(name='Yvonne').all()
        self.assertEqual(list(vals), [])

        self.sess.add(self.sm.User(name='Yvonne'))
        self.sess.flush()

        user = self.sess.query(self.sm.User).filter_by(name='Yvonne').one()
        self.assertEqual(user.name, 'Yvonne')
        self.assertIsInstance(user.id, uuid.UUID)

    def test_sqla_create_models_02(self):
        cyan = self.sm.UserGroup(
            name='cyan',
            users=[
                self.sm.User(name='Yvonne'),
                self.sm.User(name='Xander'),
            ],
        )

        self.sess.add(cyan)
        self.sess.flush()

        for name in ['Yvonne', 'Xander']:
            user = self.sess.query(self.sm.User).filter_by(name=name).one()

            self.assertEqual(user.name, name)
            self.assertEqual(user._users_UserGroup[0].name, 'cyan')
            self.assertIsInstance(user.id, uuid.UUID)

    def test_sqla_create_models_03(self):
        user0 = self.sm.User(name='Yvonne')
        user1 = self.sm.User(name='Xander')
        cyan = self.sm.UserGroup(name='cyan')

        user0._users_UserGroup.append(cyan)
        user1._users_UserGroup.append(cyan)

        self.sess.add(cyan)
        self.sess.flush()

        for name in ['Yvonne', 'Xander']:
            user = self.sess.query(self.sm.User).filter_by(name=name).one()

            self.assertEqual(user.name, name)
            self.assertEqual(user._users_UserGroup[0].name, 'cyan')
            self.assertIsInstance(user.id, uuid.UUID)

    def test_sqla_create_models_04(self):
        user = self.sm.User(name='Yvonne')
        self.sm.Post(body='this is a test', author=user)
        self.sm.Post(body='also a test', author=user)

        self.sess.add(user)
        self.sess.flush()

        res = self.sess.execute(
            select(self.sm.Post)
            .join(self.sm.Post.author)
            .where(self.sm.User.name == 'Yvonne')
        )
        self.assertEqual(
            {p.body for (p,) in res},
            {'this is a test', 'also a test'},
        )

    def test_sqla_delete_models_01(self):
        user = self.sess.query(self.sm.User).filter_by(name='Zoe').one()
        self.assertEqual(user.name, 'Zoe')
        self.assertIsInstance(user.id, uuid.UUID)

        self.sess.delete(user)
        self.sess.flush()

        vals = self.sess.query(self.sm.User).filter_by(name='Zoe').all()
        self.assertEqual(list(vals), [])

    def test_sqla_delete_models_02(self):
        post = (
            self.sess.query(self.sm.Post)
            .join(self.sm.Post.author)
            .filter(self.sm.User.name == 'Elsa')
            .one()
        )
        user_id = post.author.id

        self.sess.delete(post)
        self.sess.flush()

        vals = (
            self.sess.query(self.sm.Post)
            .join(self.sm.Post.author)
            .filter(self.sm.User.name == 'Elsa')
            .all()
        )
        self.assertEqual(list(vals), [])

        user = self.sess.get(self.sm.User, user_id)
        self.assertEqual(user.name, 'Elsa')

    def test_sqla_delete_models_03(self):
        post = (
            self.sess.query(self.sm.Post)
            .join(self.sm.Post.author)
            .filter(self.sm.User.name == 'Elsa')
            .one()
        )
        user = post.author

        self.sess.delete(post)
        self.sess.delete(user)
        self.sess.flush()

        vals = (
            self.sess.query(self.sm.Post)
            .join(self.sm.Post.author)
            .filter(self.sm.User.name == 'Elsa')
            .all()
        )
        self.assertEqual(list(vals), [])

        vals = self.sess.query(self.sm.User).filter_by(name='Elsa').all()
        self.assertEqual(list(vals), [])

    def test_sqla_delete_models_04(self):
        group = self.sess.query(
            self.sm.UserGroup).filter_by(name='green').one()
        names = {u.name for u in group.users}

        self.sess.delete(group)
        self.sess.flush()

        vals = self.sess.query(
            self.sm.UserGroup).filter_by(name='green').all()
        self.assertEqual(list(vals), [])

        users = self.sess.query(self.sm.User).all()
        for name in names:
            self.assertIn(name, {u.name for u in users})

    def test_sqla_delete_models_05(self):
        group = self.sess.query(
            self.sm.UserGroup).filter_by(name='green').one()
        for u in group.users:
            if u.name == 'Billie':
                user = u
                break

        self.sess.delete(group)
        self.sess.delete(user)
        self.sess.flush()

        vals = self.sess.query(
            self.sm.UserGroup).filter_by(name='green').all()
        self.assertEqual(list(vals), [])

        users = self.sess.query(self.sm.User).all()
        self.assertNotIn('Billie', {u.name for u in users})

    def test_sqla_update_models_01(self):
        user = self.sess.query(self.sm.User).filter_by(name='Alice').one()
        self.assertEqual(user.name, 'Alice')
        self.assertIsInstance(user.id, uuid.UUID)

        user.name = 'Xander'
        self.sess.add(user)
        self.sess.flush()

        vals = self.sess.query(self.sm.User).filter_by(name='Alice').all()
        self.assertEqual(list(vals), [])
        other = self.sess.query(self.sm.User).filter_by(name='Xander').one()
        self.assertEqual(user, other)

    def test_sqla_update_models_02(self):
        red = self.sess.query(self.sm.UserGroup).filter_by(name='red').one()
        blue = self.sess.query(self.sm.UserGroup).filter_by(name='blue').one()
        user = self.sm.User(name='Yvonne')

        self.sess.add(user)
        red.users.append(user)
        blue.users.append(user)
        self.sess.flush()

        self.assertEqual(
            {g.name for g in user._users_UserGroup},
            {'red', 'blue'},
        )
        self.assertEqual(user.name, 'Yvonne')
        self.assertIsInstance(user.id, uuid.UUID)

        group = [g for g in user._users_UserGroup if g.name == 'red'][0]
        self.assertEqual(
            {u.name for u in group.users},
            {'Alice', 'Billie', 'Cameron', 'Dana', 'Yvonne'},
        )

    def test_sqla_update_models_03(self):
        user0 = self.sess.query(self.sm.User).filter_by(name='Elsa').one()
        user1 = self.sess.query(self.sm.User).filter_by(name='Zoe').one()
        # Replace the author or a post
        post = user0._author_Post[0]
        body = post.body
        post.author = user1

        self.sess.add(post)
        self.sess.flush()

        res = self.sess.execute(
            select(self.sm.Post)
            .join(self.sm.Post.author)
            .where(self.sm.User.name == 'Zoe')
        )
        self.assertEqual(
            {p.body for (p,) in res},
            {body},
        )

    def test_sqla_update_models_04(self):
        user = self.sess.query(self.sm.User).filter_by(name='Zoe').one()
        post = (
            self.sess.query(self.sm.Post)
            .join(self.sm.Post.author)
            .filter(self.sm.User.name == 'Elsa')
            .one()
        )
        # Replace the author or a post
        post_id = post.id
        post.author = user

        self.sess.add(post)
        self.sess.flush()

        post = self.sess.get(self.sm.Post, post_id)
        self.assertEqual(post.author.name, 'Zoe')

    def test_sqla_update_models_05(self):
        # test arrays, bytes and various date/time scalars
        #
        # For the purpose of sending data creating and updating a model are
        # both testing accurate data transfer.

        res = self.sess.query(self.sm.AssortedScalars).one()

        res.name = 'New Name'
        res.vals.append('jumped')
        res.bstr = b'\x01success\x02'
        res.time = dt.time(8, 23, 54, 999_000)
        res.date = dt.date(2020, 2, 14)
        res.ts = res.ts - dt.timedelta(days=6)
        res.lts = res.lts + dt.timedelta(days=6)

        self.sess.add(res)
        self.sess.flush()

        upd = self.sess.query(self.sm.AssortedScalars).one()

        self.assertEqual(upd.name, 'New Name')
        self.assertEqual(upd.vals, ['brown', 'fox', 'jumped'])
        self.assertEqual(upd.bstr, b'\x01success\x02')
        self.assertEqual(
            upd.time,
            dt.time(8, 23, 54, 999_000),
        )
        self.assertEqual(
            upd.date,
            dt.date(2020, 2, 14),
        )
        # time zone aware
        self.assertEqual(
            upd.ts,
            dt.datetime.fromisoformat('2025-01-20T20:13:45+00:00'),
        )
        # naive datetime
        self.assertEqual(
            upd.lts,
            dt.datetime.fromisoformat('2025-02-01T20:13:45'),
        )

    def test_sqla_sorting(self):
        # Test the natural sorting function used for ordering fields, etc.

        unsorted = [
            {'name': 'zoo'},
            {'name': 'apple'},
            {'name': 'potato'},
            {'name': 'grape10'},
            {'name': 'grape1'},
            {'name': 'grape5'},
            {'name': 'grape2'},
            {'name': 'grape20'},
            {'name': 'grape25'},
            {'name': 'grape12'},
        ]

        self.assertEqual(
            list(sorted(unsorted, key=sqla.field_name_sort)),
            [
                {'name': 'apple'},
                {'name': 'grape1'},
                {'name': 'grape2'},
                {'name': 'grape5'},
                {'name': 'grape10'},
                {'name': 'grape12'},
                {'name': 'grape20'},
                {'name': 'grape25'},
                {'name': 'potato'},
                {'name': 'zoo'},
            ],
        )



from django.db.models.sql.compiler import (  # noqa
    SQLAggregateCompiler,
    SQLCompiler,
    SQLDeleteCompiler,
)
from django.db.models.sql.compiler import (  # noqa
    SQLInsertCompiler as BaseSQLInsertCompiler,
)
from django.db.models.sql.compiler import (  # noqa
    SQLUpdateCompiler as BaseSQLUpdateCompiler,
)


class GelSQLCompilerMixin:
    '''
    The reflected models have two special fields: `id` and `obj_type`. Both of
    those fields should be read-only as they are populated automatically by
    Gel and must not be modified.
    '''
    @property
    def readonly_gel_fields(self):
        try:
            # Verify that this is a Gel model reflected via Postgres protocol.
            gel_pg_meta = getattr(self.query.model, "GelPGMeta")
        except AttributeError:
            return set()
        else:
            return {'id', 'gel_type_id'}

    def as_sql(self):
        readonly_gel_fields = self.readonly_gel_fields
        if readonly_gel_fields:
            self.remove_readonly_gel_fields(readonly_gel_fields)
        return super().as_sql()


class SQLUpdateCompiler(GelSQLCompilerMixin, BaseSQLUpdateCompiler):
    def remove_readonly_gel_fields(self, names):
        '''
        Remove the values corresponding to the read-only fields.
        '''
        values = self.query.values
        # The tuple is (field, model, value)
        values[:] = (tup for tup in values if tup[0].name not in names)


class SQLInsertCompiler(GelSQLCompilerMixin, BaseSQLInsertCompiler):
    def remove_readonly_gel_fields(self, names):
        '''
        Remove the read-only fields.
        '''
        fields = self.query.fields

        try:
            fields[:] = (f for f in fields if f.name not in names)
        except AttributeError:
            # When deserializing, we might get an attribute error because this
            # list shoud be copied first:
            #
            # "AttributeError: The return type of 'local_concrete_fields'
            # should never be mutated. If you want to manipulate this list for
            # your own use, make a copy first."

            self.query.fields = [f for f in fields if f.name not in names]



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import random
from datetime import timedelta

from gel import _testbase as tb
from edgedb import errors
from edgedb.datatypes.datatypes import RelativeDuration, DateDuration


USECS_PER_HOUR = 3600000000
USECS_PER_MINUTE = 60000000
USECS_PER_SEC = 1000000


class TestDatetimeTypes(tb.SyncQueryTestCase):

    async def test_duration_01(self):

        duration_kwargs = [
            dict(),
            dict(microseconds=1),
            dict(microseconds=-1),
            dict(days=1),
            dict(days=-1),
            dict(hours=1),
            dict(seconds=-1),
            dict(microseconds=1, days=1, hours=1),
            dict(microseconds=-1, days=-1, hours=-1),
        ]

        # Fuzz it!
        for _ in range(5000):
            duration_kwargs.append(
                dict(
                    microseconds=random.randint(-1000000000, 1000000000),
                    hours=random.randint(-50, 50),
                    days=random.randint(-500, 500),
                )
            )

        durs = [timedelta(**d) for d in duration_kwargs]

        # Test encode/decode roundtrip
        durs_from_db = self.client.query('''
            WITH args := array_unpack(<array<duration>>$0)
            SELECT args;
        ''', durs)
        self.assertEqual(list(durs_from_db), durs)

    async def test_duration_02(self):
        # Make sure that when we break down the microseconds into the bigger
        # components we still get consistent values.
        tdn1h = timedelta(microseconds=-USECS_PER_HOUR)
        tdn1m = timedelta(microseconds=-USECS_PER_MINUTE)
        tdn1s = timedelta(microseconds=-USECS_PER_SEC)
        tdn1us = timedelta(microseconds=-1)
        durs = [
            (
                tdn1h, tdn1m,
                timedelta(microseconds=-USECS_PER_HOUR - USECS_PER_MINUTE),
            ),
            (
                tdn1h, tdn1s,
                timedelta(microseconds=-USECS_PER_HOUR - USECS_PER_SEC),
            ),
            (
                tdn1m, tdn1s,
                timedelta(microseconds=-USECS_PER_MINUTE - USECS_PER_SEC),
            ),
            (
                tdn1h, tdn1us,
                timedelta(microseconds=-USECS_PER_HOUR - 1),
            ),
            (
                tdn1m, tdn1us,
                timedelta(microseconds=-USECS_PER_MINUTE - 1),
            ),
            (
                tdn1s, tdn1us,
                timedelta(microseconds=-USECS_PER_SEC - 1),
            ),
        ]

        # Test encode
        durs_enc = self.client.query('''
            WITH args := array_unpack(
                <array<tuple<duration, duration, duration>>>$0)
            SELECT args.0 + args.1 = args.2;
        ''', durs)

        # Test decode
        durs_dec = self.client.query('''
            WITH args := array_unpack(
                <array<tuple<duration, duration, duration>>>$0)
            SELECT (args.0 + args.1, args.2);
        ''', durs)

        self.assertEqual(durs_enc, [True] * len(durs))
        self.assertEqual(list(durs_dec), [(d[2], d[2]) for d in durs])

    async def test_relative_duration_01(self):
        try:
            self.client.query("SELECT <cal::relative_duration>'1y'")
        except errors.InvalidReferenceError:
            self.skipTest("feature not implemented")

        delta_kwargs = [
            dict(),
            dict(microseconds=1),
            dict(microseconds=-1),
            dict(days=1),
            dict(days=-1),
            dict(months=1),
            dict(months=-1),
            dict(microseconds=1, days=1, months=1),
            dict(microseconds=-1, days=-1, months=-1),
        ]

        # Fuzz it!
        for _ in range(5000):
            delta_kwargs.append(
                dict(
                    microseconds=random.randint(-1000000000, 1000000000),
                    days=random.randint(-500, 500),
                    months=random.randint(-50, 50)
                )
            )

        durs = [RelativeDuration(**d) for d in delta_kwargs]

        # Test that RelativeDuration.__str__ formats the
        # same as <str><cal::relative_duration>
        durs_as_text = self.client.query('''
            WITH args := array_unpack(<array<cal::relative_duration>>$0)
            SELECT <str>args;
        ''', durs)

        # Test encode/decode roundtrip
        durs_from_db = self.client.query('''
            WITH args := array_unpack(<array<cal::relative_duration>>$0)
            SELECT args;
        ''', durs)

        self.assertEqual(durs_as_text, [str(d) for d in durs])
        self.assertEqual(list(durs_from_db), durs)

    async def test_relative_duration_02(self):
        d1 = RelativeDuration(microseconds=1)
        d2 = RelativeDuration(microseconds=2)
        d3 = RelativeDuration(microseconds=1)

        self.assertNotEqual(d1, d2)
        self.assertEqual(d1, d3)

        self.assertEqual(hash(d1), hash(d3))
        if hash(1) != hash(2):
            self.assertNotEqual(hash(d1), hash(d2))

        self.assertEqual(d1.days, 0)
        self.assertEqual(d1.months, 0)
        self.assertEqual(d1.microseconds, 1)

        self.assertEqual(repr(d1), '<gel.RelativeDuration "PT0.000001S">')

    async def test_relative_duration_03(self):
        # Make sure that when we break down the microseconds into the bigger
        # components we still get the sign correctly in string
        # representation.
        durs = [
            RelativeDuration(microseconds=-USECS_PER_HOUR),
            RelativeDuration(microseconds=-USECS_PER_MINUTE),
            RelativeDuration(microseconds=-USECS_PER_SEC),
            RelativeDuration(microseconds=-USECS_PER_HOUR - USECS_PER_MINUTE),
            RelativeDuration(microseconds=-USECS_PER_HOUR - USECS_PER_SEC),
            RelativeDuration(microseconds=-USECS_PER_MINUTE - USECS_PER_SEC),
            RelativeDuration(microseconds=-USECS_PER_HOUR - USECS_PER_MINUTE -
                             USECS_PER_SEC),
            RelativeDuration(microseconds=-USECS_PER_HOUR - 1),
            RelativeDuration(microseconds=-USECS_PER_MINUTE - 1),
            RelativeDuration(microseconds=-USECS_PER_SEC - 1),
            RelativeDuration(microseconds=-1),
        ]

        # Test that RelativeDuration.__str__ formats the
        # same as <str><cal::relative_duration>
        durs_as_text = self.client.query('''
            WITH args := array_unpack(<array<cal::relative_duration>>$0)
            SELECT <str>args;
        ''', durs)

        # Test encode/decode roundtrip
        durs_from_db = self.client.query('''
            WITH args := array_unpack(<array<cal::relative_duration>>$0)
            SELECT args;
        ''', durs)

        self.assertEqual(durs_as_text, [str(d) for d in durs])
        self.assertEqual(list(durs_from_db), durs)

    async def test_date_duration_01(self):
        try:
            self.client.query("SELECT <cal::date_duration>'1y'")
        except errors.InvalidReferenceError:
            self.skipTest("feature not implemented")

        delta_kwargs = [
            dict(),
            dict(days=1),
            dict(days=-1),
            dict(months=1),
            dict(months=-1),
            dict(days=1, months=1),
            dict(days=-1, months=-1),
        ]

        # Fuzz it!
        for _ in range(5000):
            delta_kwargs.append(
                dict(
                    days=random.randint(-500, 500),
                    months=random.randint(-50, 50)
                )
            )

        durs = [DateDuration(**d) for d in delta_kwargs]

        # Test that DateDuration.__str__ formats the
        # same as <str><cal::relative_duration>
        durs_as_text = self.client.query('''
            WITH args := array_unpack(<array<cal::date_duration>>$0)
            SELECT <str>args;
        ''', durs)

        # Test encode/decode roundtrip
        durs_from_db = self.client.query('''
            WITH args := array_unpack(<array<cal::date_duration>>$0)
            SELECT args;
        ''', durs)

        for db_dur, client_dur in zip(durs_as_text, durs):
            self.assertEqual(db_dur, str(client_dur))

        self.assertEqual(list(durs_from_db), durs)

    async def test_date_duration_02(self):
        # Make sure that when we break down the microseconds into the bigger
        # components we still get the sign correctly in string
        # representation.
        durs = [
            DateDuration(months=11),
            DateDuration(months=12),
            DateDuration(months=13),
            DateDuration(months=-11),
            DateDuration(months=-12),
            DateDuration(months=-13),
        ]

        # Test that DateDuration.__str__ formats the
        # same as <str><cal::date_duration>
        durs_as_text = self.client.query('''
            WITH args := array_unpack(<array<cal::date_duration>>$0)
            SELECT <str>args;
        ''', durs)

        # Test encode/decode roundtrip
        durs_from_db = self.client.query('''
            WITH args := array_unpack(<array<cal::date_duration>>$0)
            SELECT args;
        ''', durs)

        self.assertEqual(durs_as_text, [str(d) for d in durs])
        self.assertEqual(list(durs_from_db), durs)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import random

import edgedb

from gel import _testbase as tb
from edgedb import errors
from edgedb import asyncio_client


class TestAsyncIOClient(tb.AsyncQueryTestCase):
    def create_client(self, **kwargs):
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.get_database_name()
        conargs["timeout"] = 120
        conargs.update(kwargs)
        conargs.setdefault(
            "connection_class", asyncio_client.AsyncIOConnection
        )
        conargs.setdefault("max_concurrency", None)

        return tb.TestAsyncIOClient(**conargs)

    async def test_client_01(self):
        for n in {1, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                client = self.create_client(max_concurrency=10)

                async def worker():
                    self.assertEqual(await client.query_single("SELECT 1"), 1)

                tasks = [worker() for _ in range(n)]
                await asyncio.gather(*tasks)
                await client.aclose()

    async def test_client_02(self):
        for n in {1, 3, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                async with self.create_client(max_concurrency=5) as client:

                    async def worker():
                        self.assertEqual(
                            await client.query_single("SELECT 1"), 1
                        )

                    tasks = [worker() for _ in range(n)]
                    await asyncio.gather(*tasks)

    async def test_client_05(self):
        for n in {1, 3, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                client = self.create_client(max_concurrency=10)

                async def worker():
                    self.assertEqual(await client.query('SELECT 1'), [1])
                    self.assertEqual(await client.query_single('SELECT 1'), 1)
                    self.assertEqual(
                        await client.query_json('SELECT 1'), '[1]')
                    self.assertEqual(
                        await client.query_single_json('SELECT 1'), '1')

                tasks = [worker() for _ in range(n)]
                await asyncio.gather(*tasks)
                await client.aclose()

    async def test_client_transaction(self):
        client = self.create_client(max_concurrency=1)

        async for tx in client.transaction():
            async with tx:
                self.assertEqual(await tx.query_single("SELECT 7*8"), 56)

        await client.aclose()

    async def test_client_options(self):
        client = self.create_client(max_concurrency=1)

        client.with_transaction_options(
            edgedb.TransactionOptions(readonly=True))
        client.with_retry_options(
            edgedb.RetryOptions(attempts=1, backoff=edgedb.default_backoff))
        async for tx in client.transaction():
            async with tx:
                self.assertEqual(await tx.query_single("SELECT 7*8"), 56)

        await client.aclose()

    async def test_client_init_run_until_complete(self):
        client = self.create_client()
        self.assertIsInstance(client, asyncio_client.AsyncIOClient)
        await client.aclose()

    async def test_client_no_acquire_deadlock(self):
        async with self.create_client(
            max_concurrency=1,
        ) as client:

            has_sleep = await client.query_single("""
                SELECT EXISTS(
                    SELECT schema::Function FILTER .name = 'sys::_sleep'
                )
            """)
            if not has_sleep:
                self.skipTest("No sys::_sleep function")

            async def sleep_and_release():
                await client.execute("SELECT sys::_sleep(1)")

            asyncio.ensure_future(sleep_and_release())
            await asyncio.sleep(0.5)

            await client.query_single("SELECT 1")

    async def test_client_config_persistence(self):
        N = 100

        class MyConnection(asyncio_client.AsyncIOConnection):
            async def raw_query(self, query_context):
                res = await super().raw_query(query_context)
                return res + 1

        async def test(client):
            async for tx in client.transaction():
                async with tx:
                    self.assertEqual(await tx.query_single("SELECT 1"), 2)

        async with self.create_client(
            max_concurrency=10,
            connection_class=MyConnection,
        ) as client:

            await asyncio.gather(*[test(client) for _ in range(N)])

            self.assertEqual(
                sum(
                    1
                    for ch in client._impl._holders
                    if ch._con and not ch._con.is_closed()
                ),
                10,
            )

    async def test_client_connection_methods(self):
        async def test_query(client):
            i = random.randint(0, 20)
            await asyncio.sleep(random.random() / 100)
            r = await client.query("SELECT {}".format(i))
            self.assertEqual(list(r), [i])
            return 1

        async def test_query_single(client):
            i = random.randint(0, 20)
            await asyncio.sleep(random.random() / 100)
            r = await client.query_single("SELECT {}".format(i))
            self.assertEqual(r, i)
            return 1

        async def test_execute(client):
            await asyncio.sleep(random.random() / 100)
            await client.execute("SELECT {1, 2, 3, 4}")
            return 1

        async def run(N, meth):
            async with self.create_client(max_concurrency=10) as client:

                coros = [meth(client) for _ in range(N)]
                res = await asyncio.gather(*coros)
                self.assertEqual(res, [1] * N)

        methods = [
            test_query,
            test_query_single,
            test_execute,
        ]

        with tb.silence_asyncio_long_exec_warning():
            for method in methods:
                with self.subTest(method=method.__name__):
                    await run(200, method)

    async def test_client_handles_transaction_exit_in_asyncgen_1(self):
        client = self.create_client(max_concurrency=1)

        async def iterate():
            async for tx in client.transaction():
                async with tx:
                    for record in await tx.query("SELECT {1, 2, 3}"):
                        yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            agen = iterate()
            try:
                async for _ in agen:  # noqa
                    raise MyException()
            finally:
                await agen.aclose()

        await client.aclose()

    async def test_client_handles_transaction_exit_in_asyncgen_2(self):
        client = self.create_client(max_concurrency=1)

        async def iterate():
            async for tx in client.transaction():
                async with tx:
                    for record in await tx.query("SELECT {1, 2, 3}"):
                        yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            iterator = iterate()
            try:
                async for _ in iterator:  # noqa
                    raise MyException()
            finally:
                await iterator.aclose()

            del iterator

        await client.aclose()

    async def test_client_handles_asyncgen_finalization(self):
        client = self.create_client(max_concurrency=1)

        async def iterate(tx):
            for record in await tx.query("SELECT {1, 2, 3}"):
                yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            async for tx in client.transaction():
                async with tx:
                    agen = iterate(tx)
                    try:
                        async for _ in agen:  # noqa
                            raise MyException()
                    finally:
                        await agen.aclose()

        await client.aclose()

    async def test_client_close_waits_for_release(self):
        client = self.create_client(max_concurrency=1)

        flag = self.loop.create_future()
        conn_released = False

        async def worker():
            nonlocal conn_released

            async for tx in client.transaction():
                async with tx:
                    await tx.query("SELECT 42")
                    flag.set_result(True)
                    await asyncio.sleep(0.1)

            conn_released = True

        self.loop.create_task(worker())

        await flag
        await client.aclose()
        self.assertTrue(conn_released)

    async def test_client_close_timeout(self):
        client = self.create_client(max_concurrency=1)

        flag = self.loop.create_future()

        async def worker():
            async for tx in client.transaction():
                async with tx:
                    await tx.query_single("SELECT 42")
                    flag.set_result(True)
                    await asyncio.sleep(0.5)

        task = self.loop.create_task(worker())

        with self.assertRaises(asyncio.TimeoutError):
            await flag
            await asyncio.wait_for(client.aclose(), timeout=0.1)

        with self.assertRaises(errors.ClientConnectionClosedError):
            await task

    async def test_client_expire_connections(self):
        class SlowCloseConnection(asyncio_client.AsyncIOConnection):
            async def close(self, timeout=None):
                await asyncio.sleep(0.2)
                await super().close(timeout=timeout)

        client = self.create_client(
            max_concurrency=1, connection_class=SlowCloseConnection
        )

        async for tx in client.transaction():
            async with tx:
                await tx.query("SELECT 42")
                self.assertIsNotNone(client._impl._holders[0]._con)
                client._impl.expire_connections()

        self.assertIsNone(client._impl._holders[0]._con)

        await client.query("SELECT 42")
        self.assertIsNotNone(client._impl._holders[0]._con)

        client._impl.expire_connections()
        async for tx in client.transaction():
            async with tx:
                await tx.query("SELECT 42")
                with self.assertRaises(asyncio.TimeoutError):
                    await asyncio.wait_for(client.query("SELECT 42"), 1)

        await client.aclose()

    async def test_client_properties(self):
        max_concurrency = 2

        client = self.create_client(max_concurrency=max_concurrency)
        self.assertEqual(client.max_concurrency, max_concurrency)
        self.assertEqual(client.max_concurrency, max_concurrency)

        async for tx in client.transaction():
            async with tx:
                await tx.query("SELECT 42")
                self.assertEqual(client.free_size, max_concurrency - 1)

        self.assertEqual(client.free_size, max_concurrency)

        await client.aclose()

    async def _test_connection_broken(self, executor, broken_evt):
        broken_evt.set()

        with self.assertRaises(errors.ClientConnectionError):
            await executor.query_single("SELECT 123")

        broken_evt.clear()

        self.assertEqual(await executor.query_single("SELECT 123"), 123)
        broken_evt.set()
        with self.assertRaises(errors.ClientConnectionError):
            await executor.query_single("SELECT 123")
        broken_evt.clear()
        self.assertEqual(await executor.query_single("SELECT 123"), 123)
        broken_evt.set()
        broken_evt.clear()
        self.assertEqual(await executor.query_single("SELECT 123"), 123)

        tested = False
        async for tx in executor.transaction():
            async with tx:
                self.assertEqual(await tx.query_single("SELECT 123"), 123)
                if tested:
                    break
                tested = True
                broken_evt.set()
                try:
                    await tx.query_single("SELECT 123")
                except errors.ClientConnectionError:
                    broken_evt.clear()
                    raise
                else:
                    self.fail("ConnectionError not raised!")

    async def test_client_connection_broken(self):
        con_args = self.get_connect_args()
        broken = asyncio.Event()
        done = asyncio.Event()

        async def proxy(r: asyncio.StreamReader, w: asyncio.StreamWriter):
            while True:
                reader = self.loop.create_task(r.read(65536))
                waiter = self.loop.create_task(broken.wait())
                await asyncio.wait(
                    [reader, waiter],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if waiter.done():
                    reader.cancel()
                    w.close()
                    break
                else:
                    waiter.cancel()
                    data = await reader
                    if not data:
                        w.close()
                        break
                    w.write(data)

        async def cb(r: asyncio.StreamReader, w: asyncio.StreamWriter):
            ur, uw = await asyncio.open_connection(
                con_args['host'], con_args['port']
            )
            done.clear()
            task = self.loop.create_task(proxy(r, uw))
            try:
                await proxy(ur, w)
            finally:
                try:
                    await task
                finally:
                    done.set()
                    w.close()
                    uw.close()

        server = await asyncio.start_server(
            cb, '127.0.0.1', 0
        )
        port = server.sockets[0].getsockname()[1]
        client = self.create_client(
            host='127.0.0.1',
            port=port,
            max_concurrency=1,
            wait_until_available=5,
        )
        try:
            await self._test_connection_broken(client, broken)
        finally:
            await asyncio.wait_for(client.aclose(), 5)
            server.close()
            await server.wait_closed()
            broken.set()
            await done.wait()

    async def test_client_suggested_concurrency(self):
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.get_database_name()
        conargs["timeout"] = 120

        client = edgedb.create_async_client(**conargs)

        self.assertEqual(client.max_concurrency, 1)

        await client.ensure_connected()
        self.assertGreater(client.max_concurrency, 1)

        await client.aclose()

        client = edgedb.create_async_client(**conargs, max_concurrency=5)

        self.assertEqual(client.max_concurrency, 5)

        await client.ensure_connected()
        self.assertEqual(client.max_concurrency, 5)

        await client.aclose()

    def test_client_with_different_loop(self):
        conargs = self.get_connect_args()
        client = edgedb.create_async_client(**conargs)

        async def test():
            self.assertIsNot(asyncio.get_event_loop(), self.loop)
            result = await client.query_single("SELECT 42")
            self.assertEqual(result, 42)
            await asyncio.gather(
                client.query_single("SELECT 42"),
                client.query_single("SELECT 42"),
            )
            await client.aclose()

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(test())
        asyncio.set_event_loop(self.loop)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


cimport cython
cimport cpython

include "./relative_duration.pyx"
include "./date_duration.pyx"
include "./enum.pyx"
include "./config_memory.pyx"


_RecordDescriptor = EdgeRecordDesc_InitType()
Tuple = tuple
NamedTuple = EdgeNamedTuple_InitType()
Object = EdgeObject_InitType()
Record = EdgeRecord_InitType()
Set = list
Array = list

cdef str at_sign = "@"
at_sign_ptr = <cpython.PyObject*>at_sign

_EDGE_POINTER_IS_IMPLICIT = EDGE_POINTER_IS_IMPLICIT
_EDGE_POINTER_IS_LINKPROP = EDGE_POINTER_IS_LINKPROP
_EDGE_POINTER_IS_LINK = EDGE_POINTER_IS_LINK


def get_object_descriptor(obj):
    return EdgeObject_GetRecordDesc(obj)


def create_object_factory(**pointers):
    import dataclasses

    flags = ()
    names = ()
    fields = {}
    for pname, ptype in pointers.items():
        if not isinstance(ptype, set):
            ptype = {ptype}

        flag = 0
        is_linkprop = False
        for pt in ptype:
            if pt == 'link':
                flag |= EDGE_POINTER_IS_LINK
            elif pt == 'property':
                pass
            elif pt == 'link-property':
                flag |= EDGE_POINTER_IS_LINKPROP
                is_linkprop = True
            elif pt == 'implicit':
                flag |= EDGE_POINTER_IS_IMPLICIT
            else:
                raise ValueError(f'unknown pointer type {pt}')
        if is_linkprop:
            names += ("@" + pname,)
        else:
            names += (pname,)
            field = dataclasses.field()
            field.name = pname
            field._field_type = dataclasses._FIELD
            fields[pname] = field

        flags += (flag,)

    desc = EdgeRecordDesc_New(names, flags, <object>NULL)
    size = len(pointers)
    desc.set_dataclass_fields_func(lambda: fields)

    def factory(*items):
        if len(items) != size:
            raise ValueError

        o = EdgeObject_New(desc)
        for i in range(size):
            EdgeObject_SetItem(o, i, items[i])

        return o

    return factory


cdef record_desc_new(object names, object flags, object cards):
    return EdgeRecordDesc_New(names, flags, cards)


cdef record_desc_pointer_name(object desc, Py_ssize_t pos):
    return EdgeRecordDesc_PointerName(desc, pos)


cdef record_desc_pointer_card(object desc, Py_ssize_t pos):
    return EdgeRecordDesc_PointerCardinality(desc, pos)


cdef record_desc_pointer_is_link_prop(object desc, Py_ssize_t pos):
    return EdgeRecordDesc_PointerIsLinkProp(desc, pos)


cdef record_desc_pointer_is_link(object desc, Py_ssize_t pos):
    return EdgeRecordDesc_PointerIsLink(desc, pos)


cdef record_desc_pointer_is_implicit(object desc, Py_ssize_t pos):
    return EdgeRecordDesc_PointerIsImplicit(desc, pos)


cdef namedtuple_new(object namedtuple_type):
    return EdgeNamedTuple_New(namedtuple_type)


cdef namedtuple_type_new(object desc):
    return EdgeNamedTuple_Type_New(desc)


cdef object_new(object desc):
    return EdgeObject_New(desc)

cdef object_set(object obj, Py_ssize_t pos, object elem):
    EdgeObject_SetItem(obj, pos, elem)

cdef record_new(object desc):
    return EdgeRecord_New(desc)

cdef record_set(object obj, Py_ssize_t pos, object elem):
    EdgeRecord_SetItem(obj, pos, elem)
